# pylint: disable=no-self-use
import json
import os
import shutil
import sys
import time
from glob import glob
from itertools import dropwhile
from typing import Callable

import numpy as np
import pandas as pd
import pytest
import tifffile

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import AnalysisAlgorithmSelection
from PartSegCore.analysis.batch_processing import batch_backend
from PartSegCore.analysis.batch_processing.batch_backend import (
    CalculationManager,
    CalculationProcess,
    ResponseData,
    SheetData,
    do_calculation,
)
from PartSegCore.analysis.calculation_plan import (
    Calculation,
    CalculationPlan,
    CalculationTree,
    FileCalculation,
    MaskCreate,
    MaskIntersection,
    MaskSuffix,
    MaskSum,
    MaskUse,
    MeasurementCalculate,
    RootType,
    Save,
)
from PartSegCore.analysis.measurement_base import AreaType, Leaf, MeasurementEntry, Node, PerComponent
from PartSegCore.analysis.measurement_calculation import MeasurementProfile
from PartSegCore.analysis.save_functions import save_dict
from PartSegCore.image_operations import RadiusType
from PartSegCore.io_utils import LoadPlanExcel, SaveBase
from PartSegCore.json_hooks import PartSegEncoder
from PartSegCore.mask.io_functions import MaskProjectTuple, SaveROI, SaveROIOptions
from PartSegCore.mask_create import MaskProperty
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation import ROIExtractionAlgorithm, ROIExtractionResult
from PartSegCore.segmentation.noise_filtering import DimensionType
from PartSegCore.segmentation.restartable_segmentation_algorithms import LowerThresholdFlowAlgorithm
from PartSegCore.universal_const import Units
from PartSegCore.utils import BaseModel
from PartSegImage import Channel, Image, ImageWriter, TiffImageReader

ENGINE = None if pd.__version__ == "0.24.0" else "openpyxl"


class MocksCalculation:
    def __init__(self, file_path):
        self.file_path = file_path


# TODO add check of per component measurements


class DummyParams(BaseModel):
    channel: Channel = 0


class DummyExtraction(ROIExtractionAlgorithm):
    __argument_class__ = DummyParams

    @classmethod
    def support_time(cls):
        return True

    @classmethod
    def support_z(cls):  # pragma: no cover
        return True

    def calculation_run(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        channel = self.image.get_channel(0)
        if channel.max() == 0:
            raise ValueError("Empty image")
        return ROIExtractionResult(np.ones(channel.shape, dtype=np.uint8), self.get_segmentation_profile())

    def get_info_text(self):  # pragma: no cover
        return ""

    @classmethod
    def get_name(cls) -> str:
        return "Dummy"


class DummySpacingCheck(DummyExtraction):
    def calculation_run(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        assert self.image.spacing == (3, 2, 1)
        return ROIExtractionResult(np.ones(self.image.shape, dtype=np.uint8), self.get_segmentation_profile())


@pytest.fixture()
def _register_dummy_extraction():
    assert "Dummy" not in AnalysisAlgorithmSelection.__register__
    AnalysisAlgorithmSelection.register(DummyExtraction)
    yield
    AnalysisAlgorithmSelection.__register__.pop("Dummy")
    assert "Dummy" not in AnalysisAlgorithmSelection.__register__


@pytest.fixture()
def _register_dummy_spacing():
    assert "Dummy" not in AnalysisAlgorithmSelection.__register__
    AnalysisAlgorithmSelection.register(DummySpacingCheck)
    yield
    AnalysisAlgorithmSelection.__register__.pop("Dummy")
    assert "Dummy" not in AnalysisAlgorithmSelection.__register__


@pytest.fixture()
def _prepare_spacing_data(tmp_path):
    data = np.zeros((4, 1, 10, 10), dtype=np.uint8)
    data[:, :, 2:-2, 2:-2] = 1
    tifffile.imwrite(tmp_path / "test1.tiff", data)

    image = Image(data, (1, 1, 1), axes_order="ZCYX", file_path=tmp_path / "test2.tiff")
    ImageWriter.save(image, image.file_path)


@pytest.fixture()
def _prepare_mask_project_data(tmp_path):
    data = np.zeros((4, 10, 10), dtype=np.uint8)
    data[:, 2:4, 2:4] = 1
    data[:, 6:8, 2:4] = 2
    data[:, 6:8, 6:8] = 3

    image = Image(data, (1, 1, 1), axes_order="ZYX", file_path=tmp_path / "test.tiff")
    ImageWriter.save(image, image.file_path)

    roi = np.zeros(data.shape, dtype=np.uint8)
    roi[:, :5, :5] = 1
    roi[:, 5:10, :5] = 2
    roi[:, :5, 5:10] = 3
    roi[:, 5:10, 5:10] = 4

    roi = image.fit_mask_to_image(roi)

    proj = MaskProjectTuple(file_path=image.file_path, image=image, roi_info=ROIInfo(roi))

    SaveROI.save(tmp_path / "test.seg", proj, SaveROIOptions())


@pytest.fixture()
def ltww_segmentation():
    parameters = LowerThresholdFlowAlgorithm.__argument_class__(
        channel=1,
        minimum_size=200,
        threshold={
            "name": "Base/Core",
            "values": {
                "core_threshold": {"name": "Manual", "values": {"threshold": 30000}},
                "base_threshold": {"name": "Manual", "values": {"threshold": 13000}},
            },
        },
        noise_filtering={"name": "Gauss", "values": {"dimension_type": DimensionType.Layer, "radius": 1.0}},
        side_connection=False,
        flow_type={"name": "Euclidean", "values": {}},
    )

    return ROIExtractionProfile(name="test", algorithm="Lower threshold with watershed", values=parameters)


@pytest.fixture()
def measurement_list():
    chosen_fields = [
        MeasurementEntry(
            name="Segmentation Volume",
            calculation_tree=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
        ),
        MeasurementEntry(
            name="Segmentation Volume/Mask Volume",
            calculation_tree=Node(
                left=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
                op="/",
                right=Leaf(name="Volume", area=AreaType.Mask, per_component=PerComponent.No),
            ),
        ),
        MeasurementEntry(
            name="Segmentation Components Number",
            calculation_tree=Leaf(name="Components number", area=AreaType.ROI, per_component=PerComponent.No),
        ),
    ]
    statistic = MeasurementProfile(name="base_measure", chosen_fields=chosen_fields, name_prefix="")
    return MeasurementCalculate(channel=0, units=Units.µm, measurement_profile=statistic, name_prefix="")


@pytest.fixture()
def simple_measurement_list():
    chosen_fields = [
        MeasurementEntry(
            name="Segmentation Volume",
            calculation_tree=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
        ),
    ]
    statistic = MeasurementProfile(name="base_measure", chosen_fields=chosen_fields, name_prefix="")
    return MeasurementCalculate(channel=-1, units=Units.µm, measurement_profile=statistic, name_prefix="")


@pytest.fixture()
def calculation_plan_dummy(simple_measurement_list):
    tree = CalculationTree(
        RootType.Mask_project,
        [
            CalculationTree(
                ROIExtractionProfile(name="test", algorithm=DummyExtraction.get_name(), values=DummyParams()),
                [CalculationTree(simple_measurement_list, [])],
            )
        ],
    )
    return CalculationPlan(tree=tree, name="test")


@pytest.fixture()
def calculation_plan_dummy_spacing(calculation_plan_dummy):
    calculation_plan_dummy.execution_tree.operation = RootType.Image
    return calculation_plan_dummy


@pytest.fixture()
def calculation_plan(ltww_segmentation, measurement_list):
    mask_suffix = MaskSuffix(name="", suffix="_mask")
    tree = CalculationTree(
        operation=RootType.Image,
        children=[
            CalculationTree(mask_suffix, [CalculationTree(ltww_segmentation, [CalculationTree(measurement_list, [])])])
        ],
    )
    return CalculationPlan(tree=tree, name="test")


@pytest.fixture()
def calculation_plan_long(ltww_segmentation, measurement_list):
    mask_suffix = MaskSuffix(name="", suffix="_mask")
    children = []
    for i in range(20):
        measurement = measurement_list.copy()
        measurement.name_prefix = f"{i}_"
        measurement.measurement_profile = measurement.measurement_profile.copy()
        measurement.measurement_profile.name_prefix = f"{i}_"
        children.append(
            CalculationTree(mask_suffix, [CalculationTree(ltww_segmentation, [CalculationTree(measurement, [])])])
        )
    tree = CalculationTree(
        operation=RootType.Image,
        children=children,
    )
    return CalculationPlan(tree=tree, name="test")


@pytest.fixture()
def calculation_plan2(ltww_segmentation, measurement_list):
    ltww_segmentation.values.channel = 0

    tree = CalculationTree(
        RootType.Mask_project, [CalculationTree(ltww_segmentation, [CalculationTree(measurement_list, [])])]
    )
    return CalculationPlan(tree=tree, name="test2")


@pytest.fixture()
def simple_plan(simple_measurement_list):
    def _create_simple_plan(root_type: RootType, save: Save):
        parameters = {
            "channel": 0,
            "minimum_size": 200,
            "threshold": {"name": "Manual", "values": {"threshold": 13000}},
            "noise_filtering": {"name": "Gauss", "values": {"dimension_type": DimensionType.Layer, "radius": 1.0}},
            "side_connection": False,
        }
        segmentation = ROIExtractionProfile(name="test", algorithm="Lower threshold", values=parameters)
        tree = CalculationTree(
            root_type,
            [CalculationTree(segmentation, [CalculationTree(simple_measurement_list, []), CalculationTree(save, [])])],
        )
        return CalculationPlan(tree=tree, name="test")

    return _create_simple_plan


@pytest.fixture()
def calculation_plan3(ltww_segmentation):
    mask_suffix = MaskSuffix(name="", suffix="_mask")
    chosen_fields = [
        MeasurementEntry(
            name="Segmentation Volume",
            calculation_tree=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
        ),
        MeasurementEntry(
            name="Segmentation Volume/Mask Volume",
            calculation_tree=Node(
                left=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
                op="/",
                right=Leaf(name="Volume", area=AreaType.Mask, per_component=PerComponent.No),
            ),
        ),
        MeasurementEntry(
            name="Segmentation Components Number",
            calculation_tree=Leaf(name="Components number", area=AreaType.ROI, per_component=PerComponent.No),
        ),
        MeasurementEntry(
            name="Segmentation Volume per component",
            calculation_tree=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.Yes),
        ),
    ]
    statistic = MeasurementProfile(name="base_measure", chosen_fields=chosen_fields, name_prefix="")
    statistic_calculate = MeasurementCalculate(channel=0, units=Units.µm, measurement_profile=statistic, name_prefix="")
    mask_create = MaskCreate(
        name="",
        mask_property=MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=True,
            clip_to_mask=False,
            reversed_mask=False,
        ),
    )
    parameters2 = {
        "channel": 1,
        "minimum_size": 200,
        "threshold": {"name": "Manual", "values": {"threshold": 30000}},
        "noise_filtering": {"name": "Gauss", "values": {"dimension_type": DimensionType.Layer, "radius": 1.0}},
        "side_connection": False,
    }

    segmentation2 = ROIExtractionProfile(name="test", algorithm="Lower threshold", values=parameters2)
    chosen_fields = [
        MeasurementEntry(
            name="Segmentation Volume",
            calculation_tree=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
        ),
        MeasurementEntry(
            name="Segmentation Volume/Mask Volume",
            calculation_tree=Node(
                left=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.No),
                op="/",
                right=Leaf(name="Volume", area=AreaType.Mask, per_component=PerComponent.No),
            ),
        ),
        MeasurementEntry(
            name="Segmentation Components Number",
            calculation_tree=Leaf(name="Components number", area=AreaType.ROI, per_component=PerComponent.No),
        ),
        MeasurementEntry(
            name="Mask Volume per component",
            calculation_tree=Leaf(name="Volume", area=AreaType.Mask, per_component=PerComponent.Yes),
        ),
    ]
    statistic = MeasurementProfile(name="base_measure2", chosen_fields=chosen_fields[:], name_prefix="aa_")
    statistic_calculate2 = MeasurementCalculate(
        channel=0, units=Units.µm, measurement_profile=statistic, name_prefix=""
    )
    chosen_fields.append(
        MeasurementEntry(
            name="Segmentation Volume per component",
            calculation_tree=Leaf(name="Volume", area=AreaType.ROI, per_component=PerComponent.Yes),
        )
    )
    statistic = MeasurementProfile(name="base_measure3", chosen_fields=chosen_fields[:], name_prefix="bb_")
    statistic_calculate3 = MeasurementCalculate(
        channel=0, units=Units.µm, measurement_profile=statistic, name_prefix=""
    )
    tree = CalculationTree(
        RootType.Image,
        [
            CalculationTree(
                mask_suffix,
                [
                    CalculationTree(
                        ltww_segmentation,
                        [
                            CalculationTree(statistic_calculate, []),
                            CalculationTree(
                                mask_create,
                                [
                                    CalculationTree(
                                        segmentation2,
                                        [
                                            CalculationTree(statistic_calculate2, []),
                                            CalculationTree(statistic_calculate3, []),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    )
                ],
            )
        ],
    )
    return CalculationPlan(tree=tree, name="test")


@pytest.fixture(
    params=[
        MaskUse(name="test1"),
        MaskSum(name="", mask1="test1", mask2="test2"),
        MaskIntersection(name="", mask1="test1", mask2="test2"),
    ]
)
def mask_operation_plan(request, simple_measurement_list):
    parameters = {
        "channel": 0,
        "minimum_size": 200,
        "threshold": {"name": "Manual", "values": {"threshold": 13000}},
        "noise_filtering": {"name": "Gauss", "values": {"dimension_type": DimensionType.Layer, "radius": 1.0}},
        "side_connection": False,
    }
    parameters2 = dict(**parameters)
    parameters2["channel"] = 1
    segmentation = ROIExtractionProfile(name="test", algorithm="Lower threshold", values=parameters)
    segmentation2 = ROIExtractionProfile(name="test2", algorithm="Lower threshold", values=parameters2)
    tree = CalculationTree(
        RootType.Image,
        [
            CalculationTree(
                segmentation,
                [CalculationTree(MaskCreate(name="test1", mask_property=MaskProperty.simple_mask()), [])],
            ),
            CalculationTree(
                segmentation2,
                [CalculationTree(MaskCreate(name="test2", mask_property=MaskProperty.simple_mask()), [])],
            ),
            CalculationTree(
                request.param, [CalculationTree(segmentation2, [CalculationTree(simple_measurement_list, [])])]
            ),
        ],
    )
    return CalculationPlan(tree=tree, name="test")


def wait_for_calculation(manager):
    for _ in range(int(120 / 0.1)):
        res = manager.get_results()
        if res.errors and res.errors[0][0].startswith("Unknown file"):
            pytest.fail(str(res.errors))  # pragma: no cover
        if manager.has_work:
            time.sleep(0.1)
        else:
            break
    else:  # pragma: no cover
        manager.kill_jobs()
        pytest.fail("jobs hanged")

    manager.writer.finish()
    if sys.platform == "darwin":
        time.sleep(2)  # pragma: no cover
    else:
        time.sleep(0.4)


# noinspection DuplicatedCode
class TestCalculationProcess:
    def test_mask_op(self, data_test_dir, tmpdir, mask_operation_plan):
        file_path = os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif")
        calc = Calculation(
            [file_path],
            base_prefix=os.path.dirname(file_path),
            result_prefix=tmpdir,
            measurement_file_path=os.path.join(tmpdir, "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=mask_operation_plan,
            voxel_size=(1, 1, 1),
        )
        calc_process = CalculationProcess()
        res = calc_process.do_calculation(FileCalculation(file_path, calc))
        assert isinstance(res, list)
        assert isinstance(res[0], ResponseData)

    def test_one_file(self, data_test_dir, calculation_plan):
        process = CalculationProcess()
        file_path = os.path.join(data_test_dir, "stack1_components", "stack1_component5.tif")
        calc = MocksCalculation(file_path)
        process.calculation = calc
        process.image = TiffImageReader.read_image(file_path)
        process.iterate_over(calculation_plan.execution_tree)
        assert len(process.measurement[0]) == 3

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline_base(self, tmpdir, data_test_dir, monkeypatch, calculation_plan):
        monkeypatch.setattr(batch_backend, "CalculationProcess", MockCalculationProcess)
        file_pattern = os.path.join(data_test_dir, "stack1_components", "stack1_component*[0-9].tif")
        file_paths = sorted(glob(file_pattern))
        assert os.path.basename(file_paths[0]) == "stack1_component1.tif"
        calc = Calculation(
            file_paths,
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan,
            voxel_size=(1, 1, 1),
        )
        calc_process = CalculationProcess()
        for file_path in file_paths:
            res = calc_process.do_calculation(FileCalculation(file_path, calc))
            assert isinstance(res, list)
            assert isinstance(res[0], ResponseData)

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline(self, tmpdir, data_test_dir, monkeypatch, calculation_plan):
        monkeypatch.setattr(batch_backend, "CalculationProcess", MockCalculationProcess)
        file_pattern = os.path.join(data_test_dir, "stack1_components", "stack1_component*[0-9].tif")
        file_paths = sorted(glob(file_pattern))
        assert os.path.basename(file_paths[0]) == "stack1_component1.tif"
        calc = Calculation(
            file_paths,
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan,
            voxel_size=(1, 1, 1),
        )

        manager = CalculationManager()
        manager.set_number_of_workers(3)
        manager.add_calculation(calc)
        wait_for_calculation(manager)
        assert os.path.exists(os.path.join(tmpdir, "test.xlsx"))
        df = pd.read_excel(os.path.join(tmpdir, "test.xlsx"), index_col=0, header=[0, 1], engine=ENGINE)
        assert df.shape == (8, 4)
        for i in range(8):
            assert os.path.basename(df.name.units[i]) == f"stack1_component{i+1}.tif"

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline_long(self, tmpdir, data_test_dir, monkeypatch, calculation_plan_long):
        monkeypatch.setattr(batch_backend, "CalculationProcess", MockCalculationProcess)
        file_pattern = os.path.join(data_test_dir, "stack1_components", "stack1_component*[0-9].tif")
        file_paths = sorted(glob(file_pattern))[:4]
        assert os.path.basename(file_paths[0]) == "stack1_component1.tif"
        calc = Calculation(
            file_paths,
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan_long,
            voxel_size=(1, 1, 1),
        )

        manager = CalculationManager()
        manager.set_number_of_workers(3)
        manager.add_calculation(calc)
        wait_for_calculation(manager)
        res_path = os.path.join(tmpdir, "test.xlsx")
        assert os.path.exists(res_path)
        df = pd.read_excel(res_path, index_col=0, header=[0, 1], engine=ENGINE)
        assert df.shape == (4, 20 * 3 + 1)
        data, err = LoadPlanExcel.load([res_path])
        assert not err
        assert str(data["test"]) == str(calculation_plan_long)

        df2 = pd.read_excel(res_path, header=[0, 1], engine=ENGINE, sheet_name="info test")
        assert df2.shape == (152, 3)

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline_error(self, tmp_path, data_test_dir, monkeypatch, calculation_plan):
        data_dir = tmp_path / "data"
        data_dir.mkdir()
        file_pattern_copy = os.path.join(data_test_dir, "stack1_components", "stack1_component*.tif")
        file_paths = sorted(glob(file_pattern_copy))
        for el in file_paths:
            shutil.copy(el, data_dir)
        shutil.copy(data_dir / "stack1_component1.tif", data_dir / "stack1_component10.tif")
        file_pattern = os.path.join(data_dir, "stack1_component*[0-9].tif")
        file_paths = sorted(glob(file_pattern))
        result_dir = tmp_path / "result"
        result_dir.mkdir()

        assert os.path.basename(file_paths[0]) == "stack1_component1.tif"
        calc = Calculation(
            file_paths,
            base_prefix=str(data_dir),
            result_prefix=str(data_dir),
            measurement_file_path=os.path.join(result_dir, "test.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan,
            voxel_size=(1, 1, 1),
        )

        manager = CalculationManager()
        manager.set_number_of_workers(3)
        manager.add_calculation(calc)
        wait_for_calculation(manager)

        assert os.path.exists(os.path.join(result_dir, "test.xlsx"))
        df = pd.read_excel(os.path.join(result_dir, "test.xlsx"), index_col=0, header=[0, 1], engine=ENGINE)
        assert df.shape == (8, 4)
        for i in range(8):
            assert os.path.basename(df.name.units[i]) == f"stack1_component{i + 1}.tif"
        df2 = pd.read_excel(os.path.join(result_dir, "test.xlsx"), sheet_name="Errors", index_col=0, engine=ENGINE)
        assert df2.shape == (1, 2)
        str(df2.loc[0]["error description"]).startswith("[Errno 2]")

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline_mask_project(self, tmpdir, data_test_dir, calculation_plan2):
        file_pattern = os.path.join(data_test_dir, "*nucleus.seg")
        file_paths = glob(file_pattern)
        calc = Calculation(
            file_paths,
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test2.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan2,
            voxel_size=(1, 1, 1),
        )

        manager = CalculationManager()
        manager.set_number_of_workers(2)
        manager.add_calculation(calc)
        wait_for_calculation(manager)

        assert os.path.exists(os.path.join(tmpdir, "test2.xlsx"))
        df = pd.read_excel(os.path.join(tmpdir, "test2.xlsx"), index_col=0, header=[0, 1], engine=ENGINE)
        assert df.shape == (2, 4)

    def test_do_calculation(self, tmpdir, data_test_dir, calculation_plan3):
        file_path = os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif")
        calc = Calculation(
            [file_path],
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan3,
            voxel_size=(1, 1, 1),
        )
        index, res = do_calculation((1, file_path), calc)
        assert index == 1
        assert isinstance(res, list)
        assert isinstance(res[0], ResponseData)

    @pytest.mark.parametrize(
        ("file_name", "root_type"),
        [
            (os.path.join("stack1_components", "stack1_component1.tif"), RootType.Image),
            ("stack1_component1_1.tgz", RootType.Image),
            ("stack1_component1_1.tgz", RootType.Project),
            ("test_nucleus_1_1.seg", RootType.Image),
            ("test_nucleus_1_1.seg", RootType.Mask_project),
        ],
    )
    @pytest.mark.parametrize("save_method", save_dict.values())
    def test_do_calculation_save(self, tmpdir, data_test_dir, file_name, root_type, save_method: SaveBase, simple_plan):
        save_desc = Save(
            suffix="_test",
            directory="",
            algorithm=save_method.get_name(),
            short_name=save_method.get_short_name(),
            values=save_method.get_default_values(),
        )
        plan = simple_plan(root_type, save_desc)
        file_path = os.path.join(data_test_dir, file_name)
        calc = Calculation(
            [file_path],
            base_prefix=os.path.dirname(file_path),
            result_prefix=tmpdir,
            measurement_file_path=os.path.join(tmpdir, "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=plan,
            voxel_size=(1, 1, 1),
        )
        calc_process = CalculationProcess()
        res = calc_process.do_calculation(FileCalculation(file_path, calc))
        assert isinstance(res, list)
        assert isinstance(res[0], ResponseData)

    def test_do_calculation_calculation_process(self, tmpdir, data_test_dir, calculation_plan3):
        file_path = os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif")
        calc = Calculation(
            [file_path],
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan3,
            voxel_size=(1, 1, 1),
        )
        calc_process = CalculationProcess()
        res = calc_process.do_calculation(FileCalculation(file_path, calc))
        assert isinstance(res, list)
        assert isinstance(res[0], ResponseData)

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline_component_split_no_process(self, tmpdir, data_test_dir, monkeypatch, calculation_plan3):
        monkeypatch.setattr(batch_backend, "CalculationProcess", MockCalculationProcess)
        file_pattern = os.path.join(data_test_dir, "stack1_components", "stack1_component*[0-9].tif")
        file_paths = sorted(glob(file_pattern))
        assert os.path.basename(file_paths[0]) == "stack1_component1.tif"
        calc = Calculation(
            file_paths,
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan3,
            voxel_size=(1, 1, 1),
        )
        calc_process = CalculationProcess()
        for file_path in file_paths:
            res = calc_process.do_calculation(FileCalculation(file_path, calc))
            assert isinstance(res, list)
            assert isinstance(res[0], ResponseData)

    @pytest.mark.filterwarnings("ignore:This method will be removed")
    def test_full_pipeline_component_split(self, tmpdir, data_test_dir, calculation_plan3):
        file_pattern = os.path.join(data_test_dir, "stack1_components", "stack1_component*[0-9].tif")
        file_paths = glob(file_pattern)
        calc = Calculation(
            file_paths,
            base_prefix=data_test_dir,
            result_prefix=data_test_dir,
            measurement_file_path=os.path.join(tmpdir, "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan3,
            voxel_size=(1, 1, 1),
        )

        manager = CalculationManager()
        manager.set_number_of_workers(2)
        manager.add_calculation(calc)
        wait_for_calculation(manager)

        assert os.path.exists(os.path.join(tmpdir, "test3.xlsx"))
        df = pd.read_excel(os.path.join(tmpdir, "test3.xlsx"), index_col=0, header=[0, 1], engine=ENGINE)
        assert df.shape == (8, 10)
        df2 = pd.read_excel(os.path.join(tmpdir, "test3.xlsx"), sheet_name=1, index_col=0, header=[0, 1], engine=ENGINE)
        assert df2.shape[0] > 8
        assert df2.shape == (df["Segmentation Components Number"]["count"].sum(), 6)
        df3 = pd.read_excel(os.path.join(tmpdir, "test3.xlsx"), sheet_name=2, index_col=0, header=[0, 1], engine=ENGINE)
        assert df3.shape == (df["Segmentation Components Number"]["count"].sum(), 6)
        df4 = pd.read_excel(os.path.join(tmpdir, "test3.xlsx"), sheet_name=3, index_col=0, header=[0, 1], engine=ENGINE)
        assert df4.shape == (df["Segmentation Components Number"]["count"].sum(), 8)

    @pytest.mark.usefixtures("_prepare_mask_project_data")
    @pytest.mark.usefixtures("_register_dummy_extraction")
    def test_fail_single_mask_project(self, tmp_path, calculation_plan_dummy):
        file_path = str(tmp_path / "test.seg")
        calc = Calculation(
            [file_path],
            base_prefix=str(tmp_path),
            result_prefix=str(tmp_path),
            measurement_file_path=str(tmp_path / "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan_dummy,
            voxel_size=(1, 1, 1),
        )
        calc_process = CalculationProcess()
        res = calc_process.do_calculation(FileCalculation(file_path, calc))
        assert len(res) == 4
        assert sum(isinstance(x, ResponseData) for x in res) == 3
        assert isinstance(next(iter(dropwhile(lambda x: isinstance(x, ResponseData), res)))[0], ValueError)

    @pytest.mark.usefixtures("_prepare_spacing_data")
    @pytest.mark.usefixtures("_register_dummy_spacing")
    def test_spacing_overwrite(self, tmp_path, calculation_plan_dummy_spacing):
        file_path1 = str(tmp_path / "test1.tiff")
        file_path2 = str(tmp_path / "test2.tiff")
        calc = Calculation(
            [file_path1, file_path2],
            base_prefix=str(tmp_path),
            result_prefix=str(tmp_path),
            measurement_file_path=str(tmp_path / "test3.xlsx"),
            sheet_name="Sheet1",
            calculation_plan=calculation_plan_dummy_spacing,
            voxel_size=(3, 2, 1),
        )
        calc_process = CalculationProcess()
        res = calc_process.do_calculation(FileCalculation(file_path1, calc))
        assert len(res) == 1
        assert isinstance(res[0], ResponseData)
        calc.overwrite_voxel_size = True
        res = calc_process.do_calculation(FileCalculation(file_path2, calc))
        assert len(res) == 1
        assert isinstance(res[0], ResponseData)


class MockCalculationProcess(CalculationProcess):
    def do_calculation(self, calculation: FileCalculation):
        if os.path.basename(calculation.file_path) == "stack1_component1.tif":
            time.sleep(0.5)
        return super().do_calculation(calculation)


class TestSheetData:
    def test_create(self):
        cols = [("aa", "nm"), ("bb", "nm")]
        sheet_data = SheetData("test_name", cols)
        assert "test_name" in repr(sheet_data)
        assert str(cols) in repr(sheet_data)
        assert "wait_rows=0" in repr(sheet_data)

    def test_add_data(self):
        cols = [("aa", "nm"), ("bb", "nm")]
        sheet_data = SheetData("test_name", cols)
        with pytest.raises(ValueError, match="Wrong number of columns"):
            sheet_data.add_data(["aa", 1, 2, 3], None)

        with pytest.raises(ValueError, match="Wrong number of columns"):
            sheet_data.add_data(["aa", 1], None)

        sheet_data.add_data(["aa", 1, 2], None)
        assert "wait_rows=1" in repr(sheet_data)

        assert sheet_data.get_data_to_write()[0] == "test_name"
        assert "wait_rows=0" in repr(sheet_data)


def test_calculation_plan_serialize(calculation_plan_long):
    text = json.dumps(calculation_plan_long, cls=PartSegEncoder, indent=2)
    assert text.count("\n") == 7627



import numpy as np
import SimpleITK as sitk


def close_small_holes(image, max_hole_size):
    if image.dtype == bool:
        image = image.astype(np.uint8)
    if len(image.shape) == 2 or (len(image.shape) == 3 and image.shape[0] == 1):
        rev_conn = sitk.ConnectedComponent(sitk.BinaryNot(sitk.GetImageFromArray(image)), True)
        return sitk.GetArrayFromImage(sitk.BinaryNot(sitk.RelabelComponent(rev_conn, max_hole_size)))
    for layer in image:
        rev_conn = sitk.ConnectedComponent(sitk.BinaryNot(sitk.GetImageFromArray(layer)), True)
        layer[...] = sitk.GetArrayFromImage(sitk.BinaryNot(sitk.RelabelComponent(rev_conn, max_hole_size)))
    return image



# pylint: disable=no-self-use

import numpy as np

from PartSegCore.image_transforming import InterpolateImage
from PartSegCore.image_transforming.image_projection import ImageProjection, ImageProjectionParams
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image


def get_flat_image():
    data = np.zeros((1, 1, 10, 10), dtype=np.uint8)
    data[:, :, 2:-2, 2:-2] = 5
    return Image(data, (5, 5), axes_order="TZYX")


def get_cube_image():
    data = np.zeros((1, 10, 10, 10), dtype=np.uint8)
    data[:, 2:-2, 2:-2, 2:-2] = 5
    return Image(data, (10, 5, 5), axes_order="TZYX")


def get_flat_image_up():
    data = np.ones((1, 1, 10, 10), dtype=np.uint8) * 30
    data[:, :, 2:-2, 2:-2] = 10
    return Image(data, (5, 5), axes_order="TZYX")


def get_cube_image_up():
    data = np.ones((1, 10, 10, 10), dtype=np.uint8) * 30
    data[:, 2:-2, 2:-2, 2:-2] = 10
    return Image(data, (10, 5, 5), axes_order="TZYX")


class TestInterpolateImage:
    def test_simple(self):
        image = get_flat_image()
        image_res, _ = InterpolateImage.transform(image, None, {"scale": 1})
        assert np.all(image.get_data() == image_res.get_data())
        image = get_cube_image()
        image_res, _ = InterpolateImage.transform(image, None, {"scale": 1})
        assert np.all(image.get_data() == image_res.get_data())

    def test_nonzero_border(self):
        image = get_flat_image_up()
        image_res, _ = InterpolateImage.transform(image, None, {"scale": 2})
        assert image_res.get_data().min() > 0
        assert image_res.spacing == (2.5, 2.5)
        image = get_cube_image_up()
        image_res, _ = InterpolateImage.transform(image, None, {"scale": 2})
        assert image_res.get_data().min() > 0
        assert image_res.spacing == (5, 2.5, 2.5)

    def test_default_interpolate(self):
        image = get_flat_image()
        image_res, _ = InterpolateImage.transform(image, None, InterpolateImage.calculate_initial(image))
        assert image_res.spacing == (5, 5)
        assert image_res.get_data().shape == (1, 1, 1, 10, 10)
        image = get_cube_image()
        image_res, _ = InterpolateImage.transform(image, None, InterpolateImage.calculate_initial(image))
        assert image_res.spacing == (5, 5, 5)
        assert image_res.get_data().shape == (1, 1, 20, 10, 10)

    def test_multiple_interpolate(self):
        image = get_cube_image()
        image_res, _ = InterpolateImage.transform(image, None, {"scale_x": 2, "scale_y": 3, "scale_z": 4})
        assert image_res.spacing == (2.5, 5 / 3, 2.5)
        assert image_res.get_data().shape == (1, 1, 40, 30, 20)


class TestImageProjection:
    def test_basic(self):
        image = get_cube_image()
        image_res, _ = ImageProjection.transform(image, ROIInfo(None), ImageProjectionParams())
        assert image_res.shape == (1, 1, 10, 10)

    def test_no_roi(self):
        image = get_cube_image()
        image_res, roi = ImageProjection.transform(image, ROIInfo(None), ImageProjectionParams(keep_roi=True))
        assert image_res.shape == (1, 1, 10, 10)
        assert roi is None

    def test_no_mask(self):
        image = get_cube_image()
        image_res, _ = ImageProjection.transform(image, ROIInfo(None), ImageProjectionParams(keep_mask=True))
        assert image_res.shape == (1, 1, 10, 10)
        assert image_res.mask is None

    def test_cast_roi(self):
        image = get_cube_image()
        roi_arr = np.zeros((10, 10, 10), dtype=np.uint8)
        roi_arr[:, :5] = 1
        roi_arr[:, 5:] = 2
        image_res, roi_info = ImageProjection.transform(image, ROIInfo(roi_arr), ImageProjectionParams(keep_roi=True))
        assert image_res.shape == (1, 1, 10, 10)
        assert roi_info.roi.shape == (1, 1, 10, 10)
        assert np.all(roi_info.roi[0, 0] == roi_arr[0])

    def test_cast_mask(self):
        image = get_cube_image()
        mask_arr = np.zeros((10, 10, 10), dtype=np.uint8)
        mask_arr[:, :5] = 1
        mask_arr[:, 5:] = 2
        image.set_mask(mask_arr)
        image_res, _ = ImageProjection.transform(image, ROIInfo(None), ImageProjectionParams(keep_mask=True))
        assert image_res.shape == (1, 1, 10, 10)
        assert image_res.mask.shape == (1, 1, 10, 10)



import warnings

from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.analysis_utils import SegmentationPipeline, SegmentationPipelineElement
from PartSegCore.analysis.io_utils import ProjectTuple
from PartSegCore.analysis.load_functions import load_metadata
from PartSegCore.analysis.measurement_calculation import MEASUREMENT_DICT

__all__ = (
    "AnalysisAlgorithmSelection",
    "ProjectTuple",
    "SegmentationPipeline",
    "SegmentationPipelineElement",
    "MEASUREMENT_DICT",
    "load_metadata",
)


def __getattr__(name):  # pragma: no cover
    if name == "analysis_algorithm_dict":
        warnings.warn(
            "analysis_algorithm_dict is deprecated. Please use AnalysisAlgorithmSelection instead",
            category=FutureWarning,
            stacklevel=2,
        )
        return AnalysisAlgorithmSelection.__register__
    raise AttributeError(f"module {__name__} has no attribute {name}")



import logging
import multiprocessing
import os
import typing
from contextlib import suppress
from io import BytesIO
from pathlib import Path
from pickle import PicklingError  # nosec

import numpy as np
import sentry_sdk
from qtpy.QtCore import QByteArray, Qt, QTimer
from qtpy.QtGui import QIcon, QStandardItem, QStandardItemModel
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListView,
    QListWidgetItem,
    QMessageBox,
    QProgressBar,
    QPushButton,
    QSpinBox,
    QTabWidget,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)

from PartSeg import parsed_version, state_store
from PartSeg._roi_analysis.export_batch import ExportProjectDialog
from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg._roi_analysis.prepare_plan_widget import CalculatePlaner
from PartSeg.common_backend.base_settings import IO_SAVE_DIRECTORY
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.error_report import ExceptionList, ExceptionListItem
from PartSeg.common_gui.searchable_combo_box import SearchComboBox
from PartSeg.common_gui.select_multiple_files import AddFiles
from PartSeg.common_gui.universal_gui_part import Spacing, right_label
from PartSegCore.algorithm_describe_base import AlgorithmProperty
from PartSegCore.analysis.batch_processing.batch_backend import CalculationManager, get_data_loader
from PartSegCore.analysis.calculation_plan import Calculation, CalculationPlan, MaskFile
from PartSegCore.io_utils import SaveBase
from PartSegCore.segmentation.algorithm_base import SegmentationLimitException
from PartSegCore.universal_const import Units
from PartSegData import icons_dir

__author__ = "Grzegorz Bokota"


class SaveExcel(SaveBase):
    @classmethod
    def get_short_name(cls):
        return "excel"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        """empty function to satisfy interface"""

    @classmethod
    def get_name(cls) -> str:
        return "Excel (*.xlsx)"

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return []


class ProgressView(QWidget):
    """
    :type batch_manager: CalculationManager
    """

    def __init__(self, parent, batch_manager):
        super().__init__(parent)
        self.task_count = 0
        self.calculation_manager = batch_manager
        self.whole_progress = QProgressBar(self)
        self.whole_progress.setMinimum(0)
        self.whole_progress.setMaximum(1)
        self.whole_progress.setFormat("%v of %m")
        self.whole_progress.setTextVisible(True)
        self.part_progress = QProgressBar(self)
        self.part_progress.setMinimum(0)
        self.part_progress.setMaximum(1)
        self.part_progress.setFormat("%v of %m")
        self.whole_label = QLabel("All batch progress:", self)
        self.part_label = QLabel("Single batch progress:", self)
        self.cancel_remove_btn = QPushButton("Remove task")
        self.cancel_remove_btn.setDisabled(True)
        self.logs = ExceptionList(self)
        self.logs.setToolTip("Logs")
        self.task_view = QListView()
        self.task_que = QStandardItemModel(self)
        self.task_view.setModel(self.task_que)
        self.process_num_timer = QTimer()
        self.process_num_timer.setInterval(1000)
        self.process_num_timer.setSingleShot(True)
        self.process_num_timer.timeout.connect(self.change_number_of_workers)
        self.number_of_process = QSpinBox(self)
        self.number_of_process.setRange(1, multiprocessing.cpu_count())
        self.number_of_process.setValue(1)
        self.number_of_process.setToolTip("Number of process used in batch calculation")
        self.number_of_process.valueChanged.connect(self.process_num_timer_start)
        self.progress_item_dict = {}
        self.setup_ui()
        self.preview_timer = QTimer()
        self.preview_timer.setInterval(1000)
        self.preview_timer.timeout.connect(self.update_info)
        self.task_view.selectionModel().currentChanged.connect(self.task_selection_change)
        self.cancel_remove_btn.clicked.connect(self.task_cancel_remove)

    def setup_ui(self):
        layout = QGridLayout()
        layout.addWidget(self.whole_label, 0, 0, Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self.whole_progress, 0, 1, 1, 2)
        layout.addWidget(self.part_label, 1, 0, Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self.part_progress, 1, 1, 1, 2)
        lab = QLabel("Number of process:")
        lab.setToolTip("Number of process used in batch calculation")
        layout.addWidget(lab, 2, 0)
        layout.addWidget(self.number_of_process, 2, 1)
        layout.addWidget(self.logs, 3, 0, 2, 3)
        layout.addWidget(self.task_view, 0, 4, 4, 1)
        layout.addWidget(self.cancel_remove_btn, 4, 4, 1, 1)
        layout.setColumnMinimumWidth(2, 10)
        layout.setColumnStretch(2, 1)
        self.setLayout(layout)

    def task_selection_change(self, new, old):
        task: CalculationProcessItem = self.task_que.item(new.row(), new.column())
        if task is None:
            self.cancel_remove_btn.setDisabled(True)
            return
        self.cancel_remove_btn.setEnabled(True)
        if task.is_finished():
            self.cancel_remove_btn.setText(f"Remove task {task.num}")
        else:
            self.cancel_remove_btn.setText(f"Cancel task {task.num}")

    def task_cancel_remove(self):
        index = self.task_view.selectionModel().currentIndex()
        task: CalculationProcessItem = typing.cast(
            CalculationProcessItem, self.task_que.item(index.row(), index.column())
        )
        if task.is_finished():
            self.calculation_manager.remove_calculation(task.calculation)
        else:
            self.calculation_manager.cancel_calculation(task.calculation)
        self.task_que.takeRow(index.row())

    def new_task(self):
        self.whole_progress.setMaximum(self.calculation_manager.calculation_size)
        if not self.preview_timer.isActive():
            self.update_info()
            self.preview_timer.start()

    def update_info(self):
        res = self.calculation_manager.get_results()
        for el in res.errors:
            if el[0]:
                QListWidgetItem(el[0], self.logs)
            ExceptionListItem(el[1], self.logs)
            if (
                state_store.report_errors
                and parsed_version.is_devrelease
                and not isinstance(el[1][0], SegmentationLimitException)
                and isinstance(el[1][1], tuple)
            ):
                with sentry_sdk.push_scope() as scope:
                    scope.set_tag("auto_report", "true")
                    sentry_sdk.capture_event(el[1][1][0])
        self.whole_progress.setValue(res.global_counter)
        working_search = True
        for uuid, progress in res.jobs_status.items():
            calculation = self.calculation_manager.calculation_dict[uuid]
            total = len(calculation.file_list)
            if uuid in self.progress_item_dict:
                item = self.progress_item_dict[uuid]
                item.update_count(progress)
            else:
                item = CalculationProcessItem(calculation, self.task_count, progress)
                self.task_count += 1
                self.task_que.appendRow(item)
                self.progress_item_dict[uuid] = item

            if working_search and progress != total:
                self.part_progress.setMaximum(total)
                self.part_progress.setValue(progress)
                working_search = False
        if not self.calculation_manager.has_work:
            self.part_progress.setValue(self.part_progress.maximum())
            self.preview_timer.stop()
            logging.info("Progress stop")

    def process_num_timer_start(self):
        self.process_num_timer.start()

    def update_progress(self, total_progress, part_progress):
        self.whole_progress.setValue(total_progress)
        self.part_progress.setValue(part_progress)

    def set_total_size(self, size):
        self.whole_progress.setMaximum(size)

    def set_part_size(self, size):
        self.part_progress.setMaximum(size)

    def change_number_of_workers(self):
        self.calculation_manager.set_number_of_workers(self.number_of_process.value())


class FileChoose(QWidget):
    """
    :type batch_manager: CalculationManager
    """

    def __init__(self, settings: PartSettings, batch_manager, parent=None):
        super().__init__(parent)
        self.files_to_proceed = set()
        self.settings = settings
        self.batch_manager = batch_manager
        self.files_widget = AddFiles(settings, self)
        self.progress = ProgressView(self, batch_manager)
        self.run_button = QPushButton("Process")
        self.run_button.setDisabled(True)
        self.calculation_choose = SearchComboBox()
        self.calculation_choose.addItem("<no calculation>")
        self.calculation_choose.currentTextChanged.connect(self.change_situation)
        self.result_file = QLineEdit(self)
        self.result_file.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.result_file.setReadOnly(True)
        self.choose_result = QPushButton("Save result as", self)
        self.choose_result.clicked.connect(self.choose_result_file)
        self.export_data_button = QPushButton("Export batch with data", self)
        self.export_data_button.clicked.connect(self.export_data)

        self.run_button.clicked.connect(self.prepare_calculation)
        self.files_widget.file_list_changed.connect(self.change_situation)
        self.settings.batch_plans_changed.connect(self._refresh_batch_list)

        layout = QVBoxLayout()
        layout.addWidget(self.files_widget)
        calc_layout = QHBoxLayout()
        calc_layout.addWidget(QLabel("Batch workflow:"))
        calc_layout.addWidget(self.calculation_choose)
        calc_layout.addWidget(self.result_file)
        calc_layout.addWidget(self.choose_result)
        calc_layout.addWidget(self.export_data_button)
        calc_layout.addStretch()
        calc_layout.addWidget(self.run_button)
        layout.addLayout(calc_layout)
        layout.addWidget(self.progress)
        self.setLayout(layout)

        self._refresh_batch_list()

    def export_data(self):
        dialog = ExportProjectDialog(self.result_file.text(), self.files_widget.paths_input.text(), self.settings, self)
        dialog.exec_()

    def prepare_calculation(self):
        plan = self.settings.batch_plans[str(self.calculation_choose.currentText())]
        dial = CalculationPrepare(
            self.files_widget.get_paths(), plan, str(self.result_file.text()), self.settings, self.batch_manager
        )
        if dial.exec_():
            try:
                self.batch_manager.add_calculation(dial.get_data())
                self.progress.new_task()
            except PicklingError as e:  # pragma: no cover
                if state_store.develop:
                    QMessageBox.critical(self, "Pickle error", "Please restart PartSeg.")
                else:
                    raise e

    def _refresh_batch_list(self):
        current_calc = str(self.calculation_choose.currentText())
        new_list = ["<no calculation>", *sorted(n for n, p in self.settings.batch_plans.items() if not p.is_bad())]
        try:
            index = new_list.index(current_calc)
        except ValueError:
            index = 0
        self.calculation_choose.clear()
        self.calculation_choose.addItems(new_list)
        self.calculation_choose.setCurrentIndex(index)

    def change_situation(self):
        if (
            str(self.calculation_choose.currentText()) == "<no calculation>"
            or len(self.files_widget.files_to_proceed) == 0
            or not str(self.result_file.text())
        ):
            self.run_button.setDisabled(True)

        else:
            self.run_button.setEnabled(True)
        if self.calculation_choose.currentText() in self.settings.batch_plans:
            plan = self.settings.batch_plans[str(self.calculation_choose.currentText())]
            self.files_widget.mask_list = plan.get_list_file_mask()
        else:
            self.files_widget.mask_list = []

    def choose_result_file(self):
        dial = PSaveDialog(SaveExcel, system_widget=False, settings=self.settings, path=IO_SAVE_DIRECTORY)
        if dial.exec_():
            file_path = str(dial.selectedFiles()[0])
            if not os.path.splitext(file_path)[1]:
                file_path += ".xlsx"
            self.result_file.setText(file_path)
            self.change_situation()


class BatchWindow(QTabWidget):
    """
    :type settings: PartSettings
    """

    def __init__(self, settings, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Batch processing")
        self.settings = settings
        self.batch_manager = CalculationManager()
        self.file_choose = FileChoose(self.settings, self.batch_manager, self)
        self.calculate_planer = CalculatePlaner(self.settings, self)
        self.addTab(self.calculate_planer, "Prepare workflow")
        self.addTab(self.file_choose, "Input files")
        self.working = False
        with suppress(KeyError):
            geometry = self.settings.get_from_profile("batch_window_geometry")
            self.restoreGeometry(QByteArray.fromHex(bytes(geometry, "ascii")))

    def focusInEvent(self, event):
        self.calculate_planer.showEvent(event)

    def is_working(self):
        return self.working

    def terminate(self):
        self.batch_manager.writer.finish()
        self.working = False

    def closeEvent(self, event):
        if self.is_working():
            ret = QMessageBox.warning(
                self,
                "Batch work",
                "Batch work is not finished. Would you like to terminate it?",
                QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
            )
            if ret == QMessageBox.StandardButton.Yes:
                self.terminate()
            else:
                event.ignore()
        self.settings.set_in_profile("batch_window_geometry", self.saveGeometry().toHex().data().decode("ascii"))
        super().closeEvent(event)


class CalculationPrepare(QDialog):
    """
    :type mask_path_list: list[QLineEdit]
    :type mask_mapper_list: list[MaskMapper]
    """

    def __init__(
        self,
        file_list: typing.List[os.PathLike],
        calculation_plan: CalculationPlan,
        measurement_file_path: os.PathLike,
        settings: PartSettings,
        batch_manager: CalculationManager,
        parent: typing.Optional[QWidget] = None,
    ):
        """
        :param file_list: list of files to proceed
        :param calculation_plan: calculation plan for this run
        :param measurement_file_path: path to measurement result file
        :param settings: settings object
        :type batch_manager: CalculationManager
        """
        super().__init__(parent=parent)
        self.setWindowTitle("Calculation start")
        self.file_list = file_list
        self.calculation_plan = calculation_plan
        self.measurement_file_path = measurement_file_path
        self.settings = settings
        self.batch_manager = batch_manager
        self.info_label = QLabel(
            f"Information, <i><font color='{self._warning_color()}'>warnings</font></i>, "
            "<b><font color='red'>errors</font><b>"
        )
        self.voxel_size = Spacing("Voxel size", settings.image.spacing, settings.get("units_value", Units.nm))
        if len(file_list) == 1:
            all_prefix = os.path.dirname(file_list[0])
        else:
            all_prefix = os.path.commonpath(file_list)
        self.all_file_prefix = all_prefix
        self.base_prefix = QLineEdit(all_prefix, self)
        self.base_prefix.setReadOnly(True)
        self.result_prefix = QLineEdit(all_prefix, self)
        self.result_prefix.setReadOnly(True)
        self.base_prefix_btn = QPushButton("Choose data prefix")
        self.base_prefix_btn.clicked.connect(self.choose_data_prefix)
        self.result_prefix_btn = QPushButton("Choose save prefix")
        self.result_prefix_btn.clicked.connect(self.choose_result_prefix)
        self.sheet_name = QLineEdit("Sheet1")
        self.sheet_name.textChanged.connect(self.verify_data)
        self.measurement_file_path_view = QLineEdit(str(measurement_file_path))
        self.measurement_file_path_view.setReadOnly(True)

        self.overwrite_voxel_size_check = QCheckBox("Overwrite voxel size")
        self.overwrite_voxel_size_check.stateChanged.connect(self._overwrite_voxel_size_check_changed)

        self.mask_path_list = []
        self.mask_mapper_list = self.calculation_plan.get_list_file_mask()
        mask_file_list = [(i, el) for i, el in enumerate(self.mask_mapper_list) if isinstance(el, MaskFile)]

        self.state_list = np.zeros((len(self.file_list), len(self.mask_mapper_list)), dtype=np.uint8)

        self.file_list_widget = QTreeWidget()
        self.file_list_widget.header().close()

        self.execute_btn = QPushButton("Execute")
        self.execute_btn.clicked.connect(self.accept)
        self.cancel_btn = QPushButton("Cancel")
        self.cancel_btn.clicked.connect(self.close)

        self.setup_ui(mask_file_list)
        self.verify_data()

    def setup_ui(self, mask_file_list):
        mask_path_layout = QGridLayout()
        for i, (pos, mask_file) in enumerate(mask_file_list):
            if not mask_file.name:
                mask_path_layout.addWidget(right_label(f"Path to file {i + 1} with mask mapping"))
            else:
                mask_path_layout.addWidget(
                    right_label(f"Path to file {i + 1} with mask mapping for name: {mask_file.name}")
                )
            mask_path = QLineEdit(self)
            mask_path.setReadOnly(True)
            self.mask_path_list.append(mask_path)
            set_path = QPushButton("Choose file", self)
            set_path.clicked.connect(self.set_mapping_mask(i, pos))
            mask_path_layout.addWidget(mask_path, i, 1)
            mask_path_layout.addWidget(set_path, i, 2)

        layout = QGridLayout()
        layout.addWidget(self.info_label, 0, 0, 1, 5)
        layout.addWidget(self.voxel_size, 1, 0, 1, 5)
        layout.addWidget(self.overwrite_voxel_size_check, 2, 0, 1, 5)
        layout.addWidget(right_label("Measurement sheet name:"), 4, 3)
        layout.addWidget(self.sheet_name, 4, 4)
        layout.addWidget(right_label("Measurement file path:"), 3, 3)
        layout.addWidget(self.measurement_file_path_view, 3, 4)

        layout.addWidget(right_label("Data prefix:"), 3, 0)
        layout.addWidget(self.base_prefix, 3, 1)
        layout.addWidget(self.base_prefix_btn, 3, 2)
        layout.addWidget(right_label("Save prefix:"), 4, 0)
        layout.addWidget(self.result_prefix, 4, 1)
        layout.addWidget(self.result_prefix_btn, 4, 2)
        layout.addLayout(mask_path_layout, 5, 0, 1, 0)

        layout.addWidget(self.file_list_widget, 5, 0, 3, 6)
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.execute_btn)
        btn_layout.addStretch()
        btn_layout.addWidget(self.cancel_btn)
        layout.addLayout(btn_layout, 8, 0, 1, 0)
        self.setLayout(layout)

    def _warning_color(self):
        return "yellow" if self.settings.theme_name == "dark" else "blue"

    def _overwrite_voxel_size_check_changed(self):
        self.verify_data()
        if self.overwrite_voxel_size_check.isChecked():
            text = self.info_label.text()
            text += "<br><strong>Overwrite voxel size is checked. File metadata will be ignored</strong>"
            self.info_label.setText(text)

    def choose_data_prefix(self):
        dial = QFileDialog()
        dial.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        dial.setFileMode(QFileDialog.FileMode.Directory)
        dial.setDirectory(self.base_prefix.text())
        dial.setHistory(dial.history() + self.settings.get_path_history())
        if dial.exec_():
            dir_path = str(dial.selectedFiles()[0])
            self.base_prefix.setText(dir_path)

    def choose_result_prefix(self):
        dial = QFileDialog()
        dial.setOption(QFileDialog.Option.DontUseNativeDialog, True)
        dial.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        dial.setFileMode(QFileDialog.FileMode.Directory)
        dial.setDirectory(self.result_prefix.text())
        dial.setHistory(dial.history() + self.settings.get_path_history())
        if dial.exec_():
            dir_path = str(dial.selectedFiles()[0])
            self.result_prefix.setText(dir_path)

    def set_mapping_mask(self, i, pos):
        def mapping_dialog():
            dial = QFileDialog(self, "Select file")
            dial.setHistory(dial.history() + self.settings.get_path_history())
            base_path = str(self.base_prefix.text()).strip()
            if base_path:
                dial.setDirectory(base_path)
            dial.setFileMode(QFileDialog.FileMode.ExistingFile)
            if dial.exec_():
                path = str(dial.selectedFiles())
                self.mask_path_list[i].setText(path)
                file_mapper: MaskFile = self.mask_mapper_list[pos]
                file_mapper.set_map_path(path)

        return mapping_dialog

    def get_data(self):
        res = {
            "file_list": self.file_list,
            "base_prefix": str(self.base_prefix.text()),
            "result_prefix": str(self.result_prefix.text()),
            "measurement_file_path": str(self.measurement_file_path_view.text()),
            "sheet_name": str(self.sheet_name.text()),
            "calculation_plan": self.calculation_plan,
            "voxel_size": self.voxel_size.get_values(),
            "overwrite_voxel_size": self.overwrite_voxel_size_check.isChecked(),
        }
        return Calculation(**res)

    def verify_data(self):
        self.execute_btn.setEnabled(True)
        warning_color = self._warning_color()
        text = (
            f"information, <i><font color='{warning_color}'>warnings</font></i>,"
            f" <b><font color='red'>errors</font></b><br>"
            "The voxel size is for file in which metadata do not contains this information<br>"
        )
        if not self.batch_manager.is_valid_sheet_name(
            str(self.measurement_file_path_view.text()), str(self.sheet_name.text())
        ):
            text += f"<i><font color='{warning_color}'>Sheet name already in use</i></font><br>"
            self.execute_btn.setDisabled(True)
        if self.state_list.size > 0:
            val = np.unique(self.state_list)
            if 1 in val:
                self.execute_btn.setDisabled(True)
                text += f"<i><font color='{warning_color}'>Some mask map file are not set</font></i><br>"
            if 2 in val:
                self.execute_btn.setDisabled(True)
                text += "<b><font color='red'>Some mask do not exists</font><b><br>"

        if not all(os.path.exists(f) for f in self.file_list):
            self.execute_btn.setDisabled(True)
            text += "<b><font color='red'>Some files do not exists</font><b><br>"

        text = text[:-4]
        self.info_label.setText(text)

    def _check_start_conditions(self):
        for file_num, file_path in enumerate(self.file_list):
            for mask_num, mask_mapper in enumerate(self.mask_mapper_list):
                if mask_mapper.is_ready():
                    mask_path = mask_mapper.get_mask_path(file_path)
                    if os.path.exists(mask_path):
                        self.state_list[file_num, mask_num] = 0
                    else:
                        self.state_list[file_num, mask_num] = 2
                else:
                    self.state_list[file_num, mask_num] = 1
        self.verify_data()

    def showEvent(self, event):  # pragma: no cover
        super().showEvent(event)
        self._show_event_setup()

    def _show_event_setup(self):
        self._check_start_conditions()

        icon_dkt = {
            0: QIcon(os.path.join(icons_dir, "task-accepted.png")),
            1: QIcon(os.path.join(icons_dir, "task-reject.png")),
            2: QIcon(os.path.join(icons_dir, "task-attempt.png")),
        }

        text_dkt = {
            0: "Mask {} ok",
            1: "Mask {} unknown",
            2: "Mask {} file does not exists",
        }

        warn_state = np.amax(self.state_list, axis=1, initial=0)
        for file_num, file_path in enumerate(self.file_list):
            widget = QTreeWidgetItem(self.file_list_widget)
            widget.setText(0, os.path.relpath(file_path, self.all_file_prefix))
            if not os.path.exists(file_path):
                widget.setIcon(0, icon_dkt[0])
                widget.setToolTip(0, "File do not exists")
                continue
            loader, ext_match = get_data_loader(self.calculation_plan.execution_tree.operation, file_path)
            if not ext_match:
                warn_state[file_num] = 2
                widget.setToolTip(
                    0,
                    "File extension suggest that format not supported. "
                    f"Supported are {', '.join(loader.get_extensions())}",
                )
            for mask_num, mask_mapper in enumerate(self.mask_mapper_list):
                sub_widget = QTreeWidgetItem(widget)
                sub_widget.setText(0, text_dkt[self.state_list[file_num, mask_num]].format(mask_mapper.name))
                sub_widget.setIcon(0, icon_dkt[self.state_list[file_num, mask_num]])

            widget.setIcon(0, icon_dkt[warn_state[file_num]])


class CalculationProcessItem(QStandardItem):
    def __init__(self, calculation: Calculation, num: int, count, *args, **kwargs):
        text = f"Task {num} ({count}/{len(calculation.file_list)})"
        super().__init__(text, *args, **kwargs)
        self.calculation = calculation
        self.num = num
        self.count = count
        self.setToolTip(str(calculation.calculation_plan))
        self.setEditable(False)

    def update_count(self, count):
        self.count = count
        self.setText(f"Task {self.num} ({count}/{len(self.calculation.file_list)})")

    def is_finished(self) -> bool:
        return self.count == len(self.calculation.file_list)



from qtpy.QtWidgets import QComboBox, QStackedWidget, QWidget


class StackedWidgetWithSelector(QStackedWidget):
    """Stacked widget with selector which can be putted somewhere. Check show and hide event before usage"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.selector = QComboBox()
        self.selector.currentIndexChanged.connect(super().setCurrentIndex)

    # noinspection PyMethodOverriding
    def addWidget(self, widget: QWidget, name: str):
        """register widget"""
        super().addWidget(widget)
        self.selector.addItem(name)

    def setCurrentIndex(self, p: int):
        self.selector.setCurrentIndex(p)

    # noinspection PyMethodOverriding
    def insertWidget(self, p: int, widget: QWidget, name: str):
        super().insertWidget(p, widget)
        self.selector.insertItem(p, name)

    def removeWidget(self, widget: QWidget):
        index = self.indexOf(widget)
        self.selector.removeItem(index)
        super().removeWidget(widget)

    def showEvent(self, _):
        self.selector.setVisible(True)

    def hideEvent(self, _):
        self.selector.setVisible(False)



import logging
import os
import textwrap
import typing
import uuid
from abc import abstractmethod
from copy import copy, deepcopy
from enum import Enum

import local_migrator
from local_migrator import register_class, rename_key
from pydantic import BaseModel as PydanticBaseModel

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import AnalysisAlgorithmSelection
from PartSegCore.analysis.measurement_calculation import MeasurementProfile
from PartSegCore.mask_create import MaskProperty
from PartSegCore.universal_const import Units
from PartSegCore.utils import BaseModel


class MaskBase(BaseModel):
    """
    Base class for mask in calculation plan.

    :ivar str ~.name: name of mask
    """

    name: str


class RootType(Enum):
    """Defines root type which changes of data available on begin of calculation"""

    Image = 0  #: raw image
    Project = 1  #: PartSeg project with defined segmentation
    Mask_project = 2  #: Project from mask segmentation. It contains multiple elements.

    def __str__(self):
        return self.name.replace("_", " ")


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskCreate"])
class MaskCreate(MaskBase):
    """
    Description of mask creation in calculation plan.

    :ivar str ~.name: name of mask
    :ivar str ~.mask_property: instance of :py:class:`.MaskProperty`
    """

    mask_property: MaskProperty

    def __str__(self):
        return f"Mask create: {self.name}\n" + str(self.mask_property).split("\n", 1)[1]


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskUse"])
class MaskUse(MaskBase):
    """
    Reuse of already defined mask
    Will be deprecated in short time
    """


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskSum"])
class MaskSum(MaskBase):
    """
    Description of OR operation on mask

    :ivar str ~.name: name of mask
    :ivar str ~.mask1: first mask name
    :ivar str ~.mask2: second mask name
    """

    mask1: str
    mask2: str


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskIntersection"])
class MaskIntersection(MaskBase):
    """
    Description of AND operation on mask

    :ivar str ~.name: name of mask
    :ivar str ~.mask1: first mask name
    :ivar str ~.mask2: second mask name
    """

    mask1: str
    mask2: str


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.Save"])
class Save(BaseModel):
    """
    Save operation description

    :ivar str ~.suffix: suffix for saved file
    :ivar str ~.directory: name of subdirectory to save
    :ivar str ~.algorithm: name of save method
    :ivar str ~.short_name: short name of save method
    :ivar dict ~.values: parameters specific for save method
    """

    suffix: str
    directory: str
    algorithm: str
    short_name: str
    values: dict


@register_class(
    version="0.0.1",
    migrations=[("0.0.1", rename_key("statistic_profile", "measurement_profile", optional=True))],
    old_paths=["PartSeg.utils.analysis.calculation_plan.StatisticCalculate"],
)
class MeasurementCalculate(BaseModel):
    """
    Measurement calculation description

    :ivar int ~.channel: on which channel measurements should be calculated
    :ivar Units ~.units: Type of units in which results of measurements should be represented
    :ivar MeasurementProfile ~.statistic_profile: description of measurements
    :ivar str name_prefix: prefix of column names
    """

    channel: int
    units: Units
    measurement_profile: MeasurementProfile
    name_prefix: str

    @property
    def name(self):
        """name of used MeasurementProfile"""
        return self.measurement_profile.name

    def __str__(self):
        channel = "Like segmentation" if self.channel == -1 else str(self.channel)
        desc = str(self.measurement_profile).split("\n", 1)[1]
        return f"MeasurementCalculate \nChannel: {channel}\nUnits: {self.units}\n{desc}\n"


def get_save_path(op: Save, calculation: "FileCalculation") -> str:
    """
    Calculate save path base on proceeded file path and save operation parameters.
    It assume that save algorithm is registered in :py:data:`PartSegCore.analysis.save_functions.save_dict`

    :param op: operation to do
    :param calculation: information about calculation
    :return: save path
    """
    from PartSegCore.analysis.save_functions import save_dict

    extension = save_dict[op.algorithm].get_default_extension()
    rel_path = os.path.relpath(calculation.file_path, calculation.base_prefix)
    rel_path = os.path.splitext(rel_path)[0]
    if op.directory:
        file_name = os.path.basename(rel_path)
        base_rel_path = os.path.dirname(rel_path)
        return os.path.join(calculation.result_prefix, base_rel_path, op.directory, file_name + op.suffix + extension)
    return os.path.join(calculation.result_prefix, rel_path + op.suffix + extension)


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskMapper"])
class MaskMapper(BaseModel):
    """
    Base class for obtaining mask from computer disc

    :ivar ~.name: mask name
    """

    name: str

    @abstractmethod
    def get_mask_path(self, file_path: str) -> str:
        """
        Calculate mask path based od file_path

        :param file_path: path to proceeded file
        """

    @abstractmethod
    def get_parameters(self):
        """Parameters for serialize"""

    @staticmethod
    def is_ready() -> bool:
        """Check if this mask mapper can be used"""
        return True


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskSuffix"])
class MaskSuffix(MaskMapper):
    """
    Description of mask form file obtained by adding suffix to image file path

    :ivar str ~.name: mask name
    :ivar str ~.suffix: mask file path suffix
    """

    suffix: str

    def get_mask_path(self, file_path: str) -> str:
        base, ext = os.path.splitext(file_path)
        return base + self.suffix + ext

    def get_parameters(self):
        return {"name": self.name, "suffix": self.suffix}


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskSub"])
class MaskSub(MaskMapper):
    """
    Description of mask form file obtained by substitution

    :ivar str ~.name: mask name
    :ivar str ~.base: string to be searched
    :ivar str ~.repr: string to be put instead of ``base``
    """

    base: str
    rep: str

    def get_mask_path(self, file_path: str) -> str:
        dir_name, filename = os.path.split(file_path)
        filename = filename.replace(self.base, self.rep)
        return os.path.join(dir_name, filename)

    def get_parameters(self):
        return {"name": self.name, "base": self.base, "rep": self.rep}


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.MaskFile"])
class MaskFile(MaskMapper):
    # TODO Check implementation
    path_to_file: str
    name_dict: typing.Optional[dict] = None

    def is_ready(self) -> bool:
        return os.path.exists(self.path_to_file)

    def get_mask_path(self, file_path: str) -> str:
        if self.name_dict is None:
            self.parse_map()
        try:
            return self.name_dict[os.path.normpath(file_path)]
        except (KeyError, AttributeError):
            return ""

    def get_parameters(self):
        return {"name": self.name, "path_to_file": self.path_to_file}

    def set_map_path(self, value):
        self.path_to_file = value

    def parse_map(self, sep=";"):
        if not os.path.exists(self.path_to_file):  # pragma: no cover
            logging.error("File does not exists: %s", self.path_to_file)
            raise ValueError(f"File for mapping mask does not exists: {self.path_to_file}")
        with open(self.path_to_file, encoding="utf-8") as map_file:
            dir_name = os.path.dirname(self.path_to_file)
            for i, line in enumerate(map_file):
                try:
                    file_name, mask_name = line.split(sep)
                except ValueError:  # pragma: no cover
                    logging.error("Error in parsing map file\nline %s\n%s\nfrom file %s", i, line, self.path_to_file)
                    continue
                file_name = file_name.strip()
                mask_name = mask_name.strip()
                if not os.path.abspath(file_name):
                    file_name = os.path.normpath(os.path.join(dir_name, file_name))
                if not os.path.abspath(mask_name):
                    mask_name = os.path.normpath(os.path.join(dir_name, mask_name))
                self.name_dict[file_name] = mask_name


class Operations(Enum):
    """Global operations"""

    reset_to_base = 1


class PlanChanges(Enum):
    """History elements"""

    add_node = 1  #:
    remove_node = 2  #:
    replace_node = 3  #:


@register_class(old_paths=["PartSeg.utils.analysis.calculation_plan.CalculationTree"], allow_errors_in_values=True)
class CalculationTree:
    """
    Structure for describe calculation structure
    """

    def __init__(
        self,
        operation: typing.Union[PydanticBaseModel, ROIExtractionProfile, MeasurementCalculate, RootType],
        children: typing.List["CalculationTree"],
    ):
        if operation == "root":
            operation = RootType.Image
        self.operation = operation
        self.children = children

    def __str__(self):
        return f"{self.operation}:\n[{'n'.join([str(x) for x in self.children])}]"

    def __repr__(self):
        return f"CalculationTree(operation={self.operation!r}, children={self.children})"

    def as_dict(self):
        return {"operation": self.operation, "children": self.children}

    def is_bad(self):
        return any(el.is_bad() for el in self.children) or isinstance(self.operation, dict)

    def get_error_source(self):
        res = []
        for el in self.children:
            res.extend(el.get_error_source())
        if isinstance(self.operation, dict):
            res.extend(self.get_source_error_dict(self.operation))
        return res

    @classmethod
    def get_source_error_dict(cls, dkt):
        if not isinstance(dkt, dict) or "__error__" not in dkt:
            return []
        if "not found in register" in dkt["__error__"]:
            return [dkt["__error__"]]
        if "__values__" not in dkt:
            return []
        fields = local_migrator.check_for_errors_in_dkt_values(dkt["__values__"])
        res = []
        for field in fields:
            res.extend(cls.get_source_error_dict(dkt["__values__"][field]))
        return res


class NodeType(Enum):
    """Type of node in calculation"""

    segment = 1  #: segmentation
    mask = 2  #: mask creation
    measurement = 3  #: measurement calculation
    root = 4  #: root of calculation
    save = 5  #: save operation
    none = 6  #: other, like description
    file_mask = 7  #: mask load


class BaseCalculation:
    """
    Base description of calculation needed for single file

    :ivar str ~.base_prefix: path prefix which should be used to calculate relative path of processed files
    :ivar str ~.result_prefix: path prefix for saving structure
    :ivar str ~.measurement_file_path: path to file in which result of measurement should be saved
    :ivar str ~.sheet_name: name of sheet in excel file
    :ivar CalculationPlan ~.calculation_plan: plan of calculation
    :ivar str uuid: ~.uuid of whole calculation
    :ivar ~.voxel_size: default voxel size (for files which do not contains this information in metadata
    """

    def __init__(
        self,
        base_prefix: str,
        result_prefix: str,
        measurement_file_path: str,
        sheet_name: str,
        calculation_plan: "CalculationPlan",
        voxel_size: typing.Sequence[float],
        overwrite_voxel_size: bool = False,
    ):
        self.base_prefix = base_prefix
        self.result_prefix = result_prefix
        self.measurement_file_path = measurement_file_path
        self.sheet_name = sheet_name
        self.calculation_plan = calculation_plan
        self.uuid = uuid.uuid4()
        self.voxel_size = voxel_size
        self.overwrite_voxel_size = overwrite_voxel_size

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(calculation_plan={self.calculation_plan}, voxel_size={self.voxel_size}, "
            f"overwrite_voxel_size={self.overwrite_voxel_size}), "
            f"base_prefix={self.base_prefix}, result_prefix={self.base_prefix}, "
            f"measurement_file_path{self.measurement_file_path}, sheet_name={self.sheet_name})"
        )


class Calculation(BaseCalculation):
    """
    Description of whole calculation. Extended with list of all files to proceed

    :ivar str ~.base_prefix: path prefix which should be used to calculate relative path of processed files
    :ivar str ~.result_prefix: path prefix for saving structure
    :ivar str ~.measurement_file_path: path to file in which result of measurement should be saved
    :ivar str ~.sheet_name: name of sheet in excel file
    :ivar CalculationPlan ~.calculation_plan: plan of calculation
    :ivar str uuid: ~.uuid of whole calculation
    :ivar ~.voxel_size: default voxel size (for files which do not contains this information in metadata
    :ivar typing.List[str] ~.file_list: list of files to be proceed
    """

    def __init__(
        self,
        file_list,
        base_prefix,
        result_prefix,
        measurement_file_path,
        sheet_name,
        calculation_plan,
        voxel_size,
        overwrite_voxel_size=False,
    ):
        super().__init__(
            base_prefix,
            result_prefix,
            measurement_file_path,
            sheet_name,
            calculation_plan,
            voxel_size,
            overwrite_voxel_size,
        )
        self.file_list: typing.List[str] = file_list

    def get_base_calculation(self) -> BaseCalculation:
        """Extract py:class:`BaseCalculation` from instance."""
        base = BaseCalculation(
            self.base_prefix,
            self.result_prefix,
            self.measurement_file_path,
            self.sheet_name,
            self.calculation_plan,
            self.voxel_size,
            self.overwrite_voxel_size,
        )
        base.uuid = self.uuid
        return base

    @property
    def measurement(self):
        return self.calculation_plan.get_measurements()


class FileCalculation:
    """
    Description of single file calculation
    """

    def __init__(self, file_path: str, calculation: BaseCalculation):
        self.file_path = file_path
        self.calculation = calculation

    @property
    def base_prefix(self):
        """path prefix which should be used to calculate relative path of processed files"""
        return self.calculation.base_prefix

    @property
    def result_prefix(self):
        """path prefix for saving structure"""
        return self.calculation.result_prefix

    @property
    def calculation_plan(self):
        """plan of calculation"""
        return self.calculation.calculation_plan

    @property
    def uuid(self):
        """uuid of whole calculation"""
        return self.calculation.uuid

    @property
    def voxel_size(self):
        """default voxel size (for files which do not contains this information in metadata"""
        return self.calculation.voxel_size

    @property
    def overwrite_voxel_size(self):
        """overwrite voxel size"""
        return self.calculation.overwrite_voxel_size

    def __repr__(self):
        return f"FileCalculation(file_path={self.file_path}, calculation={self.calculation})"


@register_class(allow_errors_in_values=True)
class CalculationPlan:
    """
    Clean description Calculation plan.

    :type current_pos: list[int]
    :type name: str
    :type segmentation_count: int
    :type execution_tree: CalculationTree
    """

    correct_name: typing.ClassVar[typing.Dict[str, typing.Union[BaseModel, Enum]]] = {
        MaskCreate.__name__: MaskCreate,
        MaskUse.__name__: MaskUse,
        Save.__name__: Save,
        MeasurementCalculate.__name__: MeasurementCalculate,
        ROIExtractionProfile.__name__: ROIExtractionProfile,
        MaskSuffix.__name__: MaskSuffix,
        MaskSub.__name__: MaskSub,
        MaskFile.__name__: MaskFile,
        Operations.__name__: Operations,
        MaskIntersection.__name__: MaskIntersection,
        MaskSum.__name__: MaskSum,
        RootType.__name__: RootType,
    }

    def __init__(self, tree: typing.Optional[CalculationTree] = None, name: str = ""):
        if tree is None:
            self.execution_tree = CalculationTree(RootType.Image, [])
        else:
            self.execution_tree = tree
        self.segmentation_count = 0
        self.name = name
        self.current_pos = []
        self.changes = []
        self.current_node = None

    def is_bad(self):
        return self.execution_tree.is_bad()

    def get_error_source(self):
        return ", ".join(self.execution_tree.get_error_source())

    def as_dict(self):
        return {"tree": self.execution_tree, "name": self.name}

    def get_root_type(self):
        return self.execution_tree.operation

    def set_root_type(self, root_type: RootType):
        self.execution_tree.operation = root_type

    def __str__(self):
        return f"CalculationPlan<{self.name}>\n{self.execution_tree}"

    def __repr__(self):
        return f"CalculationPlan(name={self.name!r}, execution_tree={self.execution_tree!r})"

    def get_measurements(self, node: typing.Optional[CalculationTree] = None) -> typing.List[MeasurementCalculate]:
        """
        Get all measurement Calculation below given node

        :param node: Node for start, if absent then start from plan root
        :return: list of measurements
        """
        if node is None:
            node = self.execution_tree
        if isinstance(node.operation, MeasurementCalculate):
            return [node.operation]

        res = []
        for el in node.children:
            res.extend(self.get_measurements(el))
        return res

    def get_changes(self):
        ret = self.changes
        self.changes = []
        return ret

    def position(self):
        return self.current_pos

    def set_position(self, value):
        self.current_pos = value
        self.current_node = None

    def clean(self):
        self.execution_tree = CalculationTree(RootType.Image, [])
        self.current_pos = []

    def __copy__(self):
        return CalculationPlan(name=self.name, tree=deepcopy(self.execution_tree))

    def __deepcopy__(self, memo):
        return CalculationPlan(name=self.name, tree=deepcopy(self.execution_tree))

    def get_node(self, search_pos: typing.Optional[typing.List[int]] = None, parent=False) -> CalculationTree:
        """
        :param search_pos:
        :return: CalculationTree
        """
        node = self.execution_tree
        if search_pos is None:
            if self.current_node is not None and not parent:
                return self.current_node
            search_pos = self.current_pos
        if parent:
            search_pos = search_pos[:-1]
        for pos in search_pos:
            node = node.children[pos]
        return node

    def get_mask_names(self, node=None):
        """
        :type node: CalculationTree
        :param node:
        :return: set[str]
        """
        if node is None:
            node = self.get_node()
        res = set()
        if isinstance(node.operation, (MaskCreate, MaskMapper)):
            res.add(node.operation.name)
        for el in node.children:
            res |= self.get_mask_names(el)
        return res

    def get_file_mask_names(self):
        node = self.get_node()
        used_mask = self.get_reused_mask()
        tree_mask_names = self.get_mask_names(node)
        return used_mask & tree_mask_names, tree_mask_names

    @staticmethod
    def _get_reused_mask(node):
        """
        :type node: CalculationTree
        :param node:
        :return:
        """
        used_mask = set()
        for el in node.children:
            if isinstance(el.operation, MaskUse):
                used_mask.add(el.operation.name)
            elif isinstance(el.operation, (MaskSum, MaskIntersection)):
                used_mask.add(el.operation.mask1)
                used_mask.add(el.operation.mask2)
        return used_mask

    def get_reused_mask(self) -> set:
        return self._get_reused_mask(self.execution_tree)

    def get_node_type(self, parent=False) -> NodeType:
        if self.current_pos is None:
            return NodeType.none
        if not self.current_pos:
            return NodeType.root
        node = self.get_node(parent=parent)
        for klass, node_type in [
            (RootType, NodeType.root),
            ((MaskMapper, MaskIntersection, MaskSum), NodeType.file_mask),
            (MaskCreate, NodeType.mask),
            (MeasurementCalculate, NodeType.measurement),
            (ROIExtractionProfile, NodeType.segment),
            (Save, NodeType.save),
            (MaskUse, NodeType.file_mask),
        ]:
            if isinstance(node.operation, klass):
                return node_type
        if isinstance(node.operation, Operations) and node.operation == Operations.reset_to_base:
            return NodeType.mask
        raise ValueError(f"[get_node_type] unknown node type {node.operation}")

    def add_step(self, step):
        if self.current_pos is None:
            return
        node = self.get_node()
        node.children.append(CalculationTree(step, []))
        if isinstance(step, ROIExtractionProfile):
            self.segmentation_count += 1
        self.changes.append((self.current_pos, node.children[-1], PlanChanges.add_node))

    def replace_step(self, step):
        if self.current_pos is None:
            return
        node = self.get_node()
        node.operation = step
        self.changes.append((self.current_pos, node, PlanChanges.replace_node))

    def replace_name(self, name):
        if self.current_pos is None:
            return
        node = self.get_node()
        node.operation = node.operation.copy(update={"name": name})
        self.changes.append((self.current_pos, node, PlanChanges.replace_node))

    def has_children(self):
        node = self.get_node()
        return len(node.children) > 0

    def remove_step(self):
        path = copy(self.current_pos)
        if not path:
            return
        pos = path[-1]
        parent_node = self.get_node(path[:-1])
        del parent_node.children[pos]
        self.changes.append((self.current_pos, None, PlanChanges.remove_node))
        self.current_pos = self.current_pos[:-1]

    def is_segmentation(self):
        return self.segmentation_count > 0

    def set_name(self, text):
        self.name = text

    def get_execution_tree(self):
        return self.execution_tree

    def _get_save_list(self, node):
        """
        :type node: CalculationTree
        :param node:
        :return:
        """
        if isinstance(node.operation, Save):
            return [node.operation]

        res = []
        for chl in node.children:
            res.extend(self._get_save_list(chl))
        return res

    def get_save_list(self):
        return self._get_save_list(self.execution_tree)

    def get_list_file_mask(self):
        """
        :return: list[MaskMapper]
        """
        return [el.operation for el in self.execution_tree.children if isinstance(el.operation, MaskMapper)]

    def set_path_to_mapping_file(self, num, path):
        for el in self.execution_tree.children:
            if num == 0:
                el.operation.path_to_file = path
                return el.operation
            if isinstance(el.operation, MaskFile):
                num -= 1
        return None

    @classmethod
    def dict_load(cls, data_dict):
        res_plan = cls()
        name = data_dict["name"]
        res_plan.set_name(name)
        execution_tree = data_dict["execution_tree"]
        for pos, el, _ in execution_tree:
            res_plan.current_pos = pos[:-1]
            try:
                res_plan.add_step(CalculationPlan.correct_name[el["type"]](**el["values"]))
            except TypeError:
                logging.warning(el["type"])
                raise
        res_plan.changes = []
        return res_plan

    @staticmethod
    def get_el_name(el):  # noqa: C901, PLR0911, PLR0912
        """
        :param el: Plan element
        :return: str
        """
        if el.__class__.__name__ not in CalculationPlan.correct_name:
            raise ValueError(f"Unknown type {el.__class__.__name__}")
        if isinstance(el, RootType):
            return f"Root: {el}"
        if isinstance(el, Operations) and el == Operations.reset_to_base:
            return "reset project to base image with mask"
        if isinstance(el, ROIExtractionProfile):
            return f"Segmentation: {el.name}"
        if isinstance(el, MeasurementCalculate):
            if not el.name_prefix:
                return f"Measurement: {el.name}"
            return f"Measurement: {el.name} with prefix: {el.name_prefix}"
        if isinstance(el, MaskCreate):
            return f"Create mask: {el.name}" if el.name else "Create mask:"
        if isinstance(el, MaskUse):
            return f"Use mask: {el.name}"
        if isinstance(el, MaskSuffix):
            return f"File mask: {el.name} with suffix {el.suffix}"
        if isinstance(el, MaskSub):
            return f"File mask: {el.name} substitution {el.base} on {el.rep}"
        if isinstance(el, MaskFile):
            return f"File mapping mask: {el.name}, {el.path_to_file}"
        if isinstance(el, Save):
            base = el.short_name
            if el.directory:
                return f"Save {base} in directory with name {el.suffix}"
            return f"Save {base} with suffix {el.suffix}" if el.suffix else f"Save {base}"

        if isinstance(el, MaskIntersection):
            if not el.name:
                return f"Mask intersection of mask {el.mask1} and {el.mask2}"
            return f"Mask {el.name} intersection of mask {el.mask1} and {el.mask2}"
        if isinstance(el, MaskSum):
            if not el.name:
                return f"Mask sum of mask {el.mask1} and {el.mask2}"
            return f"Mask {el.name} sum of mask {el.mask1} and {el.mask2}"

        raise ValueError(f"Unknown type {type(el)}")

    def pretty_print(self) -> str:
        return f"Calculation Plan: {self.name}\n{self._pretty_print(self.execution_tree, 0)}"

    def _pretty_print(self, elem: CalculationTree, indent) -> str:
        if isinstance(elem.operation, str):
            name = elem.operation
        else:
            name = self.get_el_name(elem.operation)
        if isinstance(elem.operation, (MeasurementCalculate, ROIExtractionProfile, MaskCreate)):
            if isinstance(elem.operation, ROIExtractionProfile):
                txt = elem.operation.pretty_print(AnalysisAlgorithmSelection)
            else:
                txt = str(elem.operation)
            txt = "\n".join(txt.split("\n")[1:])
            name += "\n" + textwrap.indent(txt, " " * (indent + 4))

        if elem.children:
            suffix = "\n" + "\n".join(self._pretty_print(x, indent + 2) for x in elem.children)

        else:
            suffix = ""

        return " " * indent + name + suffix



import json
import logging
import os
import tarfile
import time
import typing
import zipfile
from contextlib import suppress
from pathlib import Path

import requests
from openpyxl import load_workbook
from qtpy.QtCore import QByteArray, Qt
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QFormLayout,
    QGridLayout,
    QLabel,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QTextEdit,
    QTreeWidget,
    QTreeWidgetItem,
    QWidget,
)
from superqt import QCollapsible
from superqt.utils import thread_worker

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_backend.base_settings import IO_SAVE_DIRECTORY
from PartSeg.common_gui.custom_load_dialog import PLoadDialog, SelectDirectoryDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.main_window import OPEN_DIRECTORY
from PartSeg.common_gui.select_multiple_files import IO_BATCH_DIRECTORY
from PartSegCore.io_utils import LoadPlanExcel
from PartSegData import icons_dir

REQUESTS_TIMEOUT = 600

NO_FILES = "No files to export"
MISSING_FILES = "Some files do not exists"


class ExportProjectDialog(QDialog):
    """Export data for zenodo"""

    def __init__(
        self, excel_path: str, base_folder: str, settings: PartSettings, parent: typing.Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)
        self.setWindowTitle("Export batch with data")
        self.settings = settings
        self._all_files_exists = False
        self.info_label = QLabel()
        self.info_label.setVisible(False)
        self.info_label.setWordWrap(True)
        self.info_label.setTextFormat(getattr(Qt.TextFormat, "MarkdownText", Qt.TextFormat.AutoText))
        self.info_label.setOpenExternalLinks(True)
        self.excel_path = QLineEdit(excel_path)
        self.base_folder = QLineEdit(base_folder)
        self.zenodo_token = QLineEdit(settings.get("zenodo_token", ""))
        self.zenodo_token.setToolTip(
            "You can get token from https://zenodo.org/account/settings/applications/."
            " The token for sandbox and production are different"
        )
        self.zenodo_title = QLineEdit()
        self.zenodo_author = QLineEdit(settings.get("zenodo_author", ""))
        self.zenodo_author.setToolTip(
            "Only first author could be used from this widget. Other you need to add manually"
        )
        self.zenodo_affiliation = QLineEdit(settings.get("zenodo_affiliation", ""))
        self.zenodo_orcid = QLineEdit(settings.get("zenodo_orcid", ""))
        self.zenodo_description = QTextEdit()
        self.zenodo_description.setPlaceholderText("Put your dataset description here")
        self.zenodo_description.setSizeAdjustPolicy(QTextEdit.SizeAdjustPolicy.AdjustToContents)
        self.zenodo_sandbox = QCheckBox()
        self.zenodo_sandbox.setToolTip("Use https://sandbox.zenodo.org instead of https://zenodo.org")
        self.excel_path_btn = QPushButton("Select excel file")
        self.base_folder_btn = QPushButton("Select base folder")
        self.info_box = QTreeWidget()
        self.info_box.header().close()
        self.zenodo_collapse = QCollapsible("Zenodo export")

        self.export_btn = QPushButton("Export")
        self.export_btn.setDisabled(True)
        self.export_to_zenodo_btn = QPushButton("Export to zenodo")
        self.export_to_zenodo_btn.setDisabled(True)
        self.progress_bar = QProgressBar()
        self.progress_bar.setVisible(False)
        self.worker = None

        self.zenodo_token.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.excel_path.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.base_folder.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.zenodo_title.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.zenodo_author.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.zenodo_affiliation.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.zenodo_description.textChanged.connect(self._check_if_enable_zenodo_btn)
        self.excel_path.textChanged.connect(self._excel_path_changed)
        self.base_folder.textChanged.connect(self._excel_path_changed)
        self.base_folder_btn.clicked.connect(self.select_folder)
        self.excel_path_btn.clicked.connect(self.select_excel)
        self.export_btn.clicked.connect(self._export_archive)
        self.export_to_zenodo_btn.clicked.connect(self._export_to_zenodo)

        self._setup_ui()

    def _setup_ui(self):
        layout = QGridLayout()

        layout.addWidget(self.info_label, 0, 0, 1, 3)
        layout.addWidget(QLabel("Excel file"), 1, 0)
        layout.addWidget(self.excel_path, 1, 1)
        layout.addWidget(self.excel_path_btn, 1, 2)
        layout.addWidget(QLabel("Base folder"), 2, 0)
        layout.addWidget(self.base_folder, 2, 1)
        layout.addWidget(self.base_folder_btn, 2, 2)
        layout.addWidget(self.zenodo_collapse, 3, 0, 1, 3)
        layout.addWidget(self.info_box, 4, 0, 1, 3)
        layout.addWidget(self.progress_bar, 5, 0, 1, 3)

        layout.addWidget(self.export_btn, 6, 0)
        layout.addWidget(self.export_to_zenodo_btn, 6, 2)

        self.setLayout(layout)

        zenodo_layout = QFormLayout()
        zenodo_layout.addRow("Zenodo token", self.zenodo_token)
        zenodo_layout.addRow("Dataset title", self.zenodo_title)
        zenodo_layout.addRow("Dataset author", self.zenodo_author)
        zenodo_layout.addRow("Affiliation", self.zenodo_affiliation)
        zenodo_layout.addRow("ORCID", self.zenodo_orcid)
        zenodo_layout.addRow(QLabel("Description:"))
        zenodo_layout.addRow(self.zenodo_description)
        sandbox_label = QLabel("Upload to sandbox")
        sandbox_label.setToolTip(self.zenodo_sandbox.toolTip())
        zenodo_layout.addRow(sandbox_label, self.zenodo_sandbox)

        widg = QWidget()
        widg.setLayout(zenodo_layout)
        self.zenodo_collapse.addWidget(widg)
        with suppress(KeyError):
            geometry = self.settings.get_from_profile("export_window_geometry")
            self.restoreGeometry(QByteArray.fromHex(bytes(geometry, "ascii")))

    def closeEvent(self, event) -> None:
        self.settings.set_in_profile("export_window_geometry", self.saveGeometry().toHex().data().decode("ascii"))
        super().closeEvent(event)

    def _export_archive(self):
        self.info_label.setVisible(False)
        self.export_btn.setDisabled(True)
        self.export_to_zenodo_btn.setDisabled(True)
        dlg = PSaveDialog(
            "Archive name (*.tgz *.zip *.tbz2 *.txy)",
            settings=self.settings,
            path=IO_SAVE_DIRECTORY,
            parent=self,
        )
        if dlg.exec_():
            self.progress_bar.setVisible(True)
            self.progress_bar.setRange(0, self.info_box.topLevelItemCount() + 1)
            self.progress_bar.setValue(0)
            export_to_archive_ = thread_worker(
                export_to_archive,
                connect={
                    "yielded": self._progress,
                    "finished": self._export_finished,
                    "errored": self._export_errored,
                    "returned": self._export_returned,
                },
                start_thread=False,
            )
            self.worker = export_to_archive_(
                excel_path=Path(self.excel_path.text()),
                base_folder=Path(self.base_folder.text()),
                target_path=Path(dlg.selectedFiles()[0]),
            )
            self.worker.start()

    def _export_finished(self):
        self.progress_bar.setVisible(False)
        self.progress_bar.setValue(0)
        self.worker = None
        self._check_if_enable_zenodo_btn()
        self._check_if_enable_export_btn()

    def _export_returned(self, result):
        self.info_label.setText(f"Export finished. {result}")
        self.info_label.setVisible(True)

    def _progress(self, value: int):
        self.progress_bar.setValue(value)

    def _export_to_zenodo(self):
        self.info_label.setVisible(False)
        self.export_btn.setDisabled(True)
        self.export_to_zenodo_btn.setDisabled(True)
        self.progress_bar.setVisible(True)
        self.progress_bar.setRange(0, self.info_box.topLevelItemCount() + 1)
        self.progress_bar.setValue(0)

        export_to_zenodo_ = thread_worker(
            export_to_zenodo,
            connect={
                "yielded": self._progress,
                "finished": self._export_finished,
                "errored": self._export_errored,
                "returned": self._export_returned,
            },
            start_thread=False,
        )
        url = "https://zenodo.org/api/deposit/depositions"
        if self.zenodo_sandbox.isChecked():
            url = "https://sandbox.zenodo.org/api/deposit/depositions"
        self.worker = export_to_zenodo_(
            excel_path=Path(self.excel_path.text()),
            base_folder=Path(self.base_folder.text()),
            title=self.zenodo_title.text(),
            author=self.zenodo_author.text(),
            affiliation=self.zenodo_affiliation.text(),
            description=self.zenodo_description.toPlainText(),
            zenodo_token=self.zenodo_token.text(),
            orcid=self.zenodo_orcid.text(),
            zenodo_url=url,
        )
        self.settings.set("zenodo_token", self.zenodo_token.text())
        self.settings.set("zenodo_author", self.zenodo_author.text())
        self.settings.set("zenodo_affiliation", self.zenodo_affiliation.text())
        self.settings.set("zenodo_orcid", self.zenodo_orcid.text())
        self.worker.start()

    def _export_errored(self, value):
        self.info_label.setText(f"Error: {value}")
        self.info_label.setVisible(True)

    def _could_export(self):
        dir_path = Path(self.base_folder.text())
        excel_path = Path(self.excel_path.text())
        return self._all_files_exists and dir_path.is_dir() and excel_path.exists() and excel_path.is_file()

    def _check_if_enable_zenodo_btn(self):
        self.export_to_zenodo_btn.setEnabled(
            bool(
                self._could_export()
                and len(self.zenodo_token.text()) > 5
                and self.zenodo_author.text()
                and self.zenodo_affiliation.text()
                and self.zenodo_title.text()
                and self.zenodo_description.toPlainText()
            )
        )

    def _check_if_enable_export_btn(self):
        self.export_btn.setEnabled(self._could_export())

    def select_folder(self):
        dial = SelectDirectoryDialog(
            settings=self.settings,
            settings_path=[IO_BATCH_DIRECTORY, OPEN_DIRECTORY],
            default_directory=str(Path.home()),
            parent=self,
        )
        if dial.exec_():
            self.base_folder.setText(dial.selectedFiles()[0])

    def select_excel(self):
        dial = PLoadDialog(LoadPlanExcel, settings=self.settings, path=IO_SAVE_DIRECTORY)
        if dial.exec_():
            file_path = str(dial.selectedFiles()[0])
            if not os.path.splitext(file_path)[1]:
                file_path += ".xlsx"
            self.excel_path.setText(file_path)

    def _excel_path_changed(self):
        excel_path = Path(self.excel_path.text())
        if not excel_path.exists() or not excel_path.is_file():
            return
        not_icon = QIcon(os.path.join(icons_dir, "task-reject.png"))
        ok_icon = QIcon(os.path.join(icons_dir, "task-accepted.png"))

        file_and_presence_list = _extract_information_from_excel_to_export(excel_path, self.base_folder.text())
        self.info_box.clear()
        presence_all = bool(file_and_presence_list)
        if not presence_all:
            self.info_label.setText(NO_FILES)
            self.info_label.setVisible(True)
        else:
            self.info_label.setText("")
            self.info_label.setVisible(False)
            for file_path, presence in file_and_presence_list:
                widget = QTreeWidgetItem(self.info_box)
                widget.setText(0, file_path)
                if not presence:
                    widget.setIcon(0, not_icon)
                    widget.setToolTip(0, "File do not exists")
                else:
                    widget.setIcon(0, ok_icon)
                presence_all &= presence

        self._all_files_exists = presence_all
        self._check_if_enable_zenodo_btn()
        self._check_if_enable_export_btn()


def _extract_information_from_excel_to_export(
    excel_path: typing.Union[str, Path], base_folder: typing.Union[str, Path]
) -> typing.List[typing.Tuple[str, bool]]:
    """Extract information from Excel file to export"""
    file_list = []
    file_set = set()
    base_folder = Path(base_folder)

    xlsx = load_workbook(filename=excel_path, read_only=True)
    for sheet in xlsx.worksheets:
        if sheet.cell(1, 2).value != "name":
            continue
        index = 4  # offset
        while image_path := sheet.cell(index, 2).value:
            index += 1
            if image_path in file_set:
                continue
            file_set.add(image_path)
            file_list.append((image_path, (base_folder / image_path).exists()))

    return file_list


def export_to_archive(excel_path: Path, base_folder: Path, target_path: Path):
    """
    Export files to archive

    :param Path excel_path: path to Excel file
    :param Path base_folder: base folder from where paths are calculated
    :param Path target_path: path to archive
    """

    file_list = _extract_information_from_excel_to_export(excel_path, base_folder)
    if not file_list:
        raise ValueError(NO_FILES)
    if not all(presence for _, presence in file_list):
        raise ValueError(MISSING_FILES)
    ext = target_path.suffix
    if ext == ".zip":
        with zipfile.ZipFile(target_path, "w") as zip_file:
            zip_file.write(excel_path, arcname=excel_path.name)
            yield 1
            for i, (file_path, _) in enumerate(file_list, start=2):
                zip_file.write(base_folder / file_path, arcname=file_path)
                yield i

        return target_path

    mode_dict = {
        ".tgz": "w:gz",
        ".gz": "w:gz",
        ".tbz2": "w:bz2",
        ".bz2": "w:bz2",
        ".txz": "w:xz",
        ".xz": "w:xz",
        ".tar": "w:",
    }

    mode = mode_dict.get(ext)
    if mode is None:
        raise ValueError("Unknown archive type")

    with tarfile.open(target_path, mode=mode) as tar:
        tar.add(excel_path, arcname=excel_path.name)
        yield 1
        for i, (file_path, _) in enumerate(file_list, start=2):
            tar.add(base_folder / file_path, arcname=file_path)
            yield i

    return target_path


class ZenodoCreateError(ValueError):
    pass


def sleep_with_rate(response: requests.Response):
    """Sleep with rate limit"""
    if "X-RateLimit-Remaining" not in response.headers:
        return
    remaining = int(response.headers["X-RateLimit-Remaining"])
    if remaining > 0:
        return
    reset = int(response.headers["X-RateLimit-Reset"])
    sleep_time = reset - time.time()
    if sleep_time > 0:
        logging.info("Sleeping for {sleep_time} seconds", extra={"sleep_time": sleep_time})
        time.sleep(sleep_time)


def export_to_zenodo(
    excel_path: Path,
    base_folder: Path,
    zenodo_token: str,
    title: str,
    author: str,
    affiliation: str,
    orcid: str,
    description: str,
    zenodo_url: str = "https://zenodo.org/api/deposit/depositions",
):
    """
    Export project to Zenodo

    :param excel_path: path to Excel file with output from batch processing
    :param base_folder: base folder from where paths are calculated
    :param zenodo_token: Zenodo API token
    :param title: title of the deposition
    :param author: author of the deposition
    :param affiliation: affiliation of the author
    :param orcid: ORCID of the author
    :param description: description of the deposition
    :param zenodo_url: Zenodo API URL
    :return:
    """
    file_list = _extract_information_from_excel_to_export(excel_path, base_folder)
    if not file_list:
        raise ValueError(NO_FILES)
    if not all(presence for _, presence in file_list):
        raise ValueError(MISSING_FILES)
    params = {"access_token": zenodo_token}
    headers = {"Content-Type": "application/json"}
    initial_request = requests.post(
        zenodo_url,
        params=params,
        json={},
        headers=headers,
        timeout=REQUESTS_TIMEOUT,
    )
    if initial_request.status_code != 201:
        raise ZenodoCreateError(
            "Can't create deposition. Please check your zenodo token."
            " Please remember that token for sandbox and production are different. "
            f" You could create token at "
            f"{zenodo_url.replace('api/deposit/depositions', 'account/settings/applications/')}"
            f" {initial_request.status_code} {initial_request.json()['message']}"
        )
    bucket_url = initial_request.json()["links"]["bucket"]
    deposition_id = initial_request.json()["id"]
    deposit_url = initial_request.json()["links"]["html"]

    data = {
        "metadata": {
            "title": title,
            "upload_type": "dataset",
            "creators": [{"name": author, "affiliation": affiliation}],
        }
    }

    if description:
        data["metadata"]["description"] = description
    if orcid:
        data["metadata"]["creators"][0]["orcid"] = orcid

    r = requests.put(
        f"{zenodo_url}/{deposition_id}",
        params=params,
        data=json.dumps(data),
        headers=headers,
        timeout=REQUESTS_TIMEOUT,
    )
    if r.status_code != 200:
        raise ZenodoCreateError(f"Can't update deposition metadata. Status code: {r.status_code}")

    with excel_path.open(mode="rb") as fp:
        r = requests.put(
            f"{bucket_url}/{excel_path.name}",
            data=fp,
            params=params,
            timeout=REQUESTS_TIMEOUT,
        )
        if r.status_code != 200:
            raise ZenodoCreateError(f"Can't upload excel file. Status code: {r.status_code}")
        yield 1

    for i, (filename, _) in enumerate(file_list, start=2):
        sleep_with_rate(r)
        with (base_folder / filename).open(mode="rb") as fp:
            r = requests.put(
                f"{bucket_url}/{filename}",
                data=fp,
                params=params,
                timeout=REQUESTS_TIMEOUT,
            )
            if r.status_code != 200:
                raise ZenodoCreateError(f"Can't upload file {filename}. Status code: {r.status_code}")
            yield i

    return deposit_url



import re
import typing

import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QButtonGroup,
    QDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidgetItem,
    QPushButton,
    QRadioButton,
    QTextEdit,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)

from PartSeg.common_gui.searchable_list_widget import SearchableListWidget
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection


class ObjectPreviewProtocol(typing.Protocol):
    def preview_object(self, ob):
        raise NotImplementedError


class ObjectPreview(QTextEdit):
    """Base class for viewer used by :py:class:`ExportDialog` to preview data"""

    def preview_object(self, ob):
        raise NotImplementedError


class StringViewer(ObjectPreview):
    """Simple __str__ serialization"""

    def preview_object(self, ob):
        self.setText(str(ob))


class ProfileDictViewer(ObjectPreview):
    """
    Preview of :py:class:`SegmentationProfile`.
    Serialized using :py:meth:`ObjectPreview.pretty_print`.
    """

    def preview_object(self, ob: ROIExtractionProfile):
        text = ob.pretty_print(AnalysisAlgorithmSelection.__register__)
        self.setText(text)


class ExportDialog(QDialog):
    def __init__(self, export_dict, viewer, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle("Export")
        self.export_dict = export_dict
        self.viewer = viewer()
        self.list_view = SearchableListWidget()
        self.check_state = np.zeros(len(export_dict), dtype=bool)
        self.check_state[...] = True
        for el in sorted(export_dict.keys()):
            item = QListWidgetItem(el)
            # noinspection PyTypeChecker
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Checked)
            self.list_view.addItem(item)

        self.checked_num = len(export_dict)

        self.export_btn = QPushButton("Export")
        self.cancel_btn = QPushButton("Cancel")
        self.check_btn = QPushButton("Check all")
        self.uncheck_btn = QPushButton("Uncheck all")

        self.cancel_btn.clicked.connect(self.close)
        self.export_btn.clicked.connect(self.accept)
        self.check_btn.clicked.connect(self.check_all)
        self.uncheck_btn.clicked.connect(self.uncheck_all)

        self.list_view.itemSelectionChanged.connect(self.preview)
        self.list_view.itemChanged.connect(self.checked_change)

        layout = QVBoxLayout()
        info_layout = QHBoxLayout()
        info_layout.addWidget(self.list_view)
        info_layout.addWidget(self.viewer)
        layout.addLayout(info_layout)
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.check_btn)
        btn_layout.addWidget(self.uncheck_btn)
        btn_layout.addStretch()
        btn_layout.addWidget(self.export_btn)
        btn_layout.addWidget(self.cancel_btn)
        layout.addLayout(btn_layout)
        self.setLayout(layout)

    def checked_change(self, item):
        if item.checkState() == Qt.CheckState.Unchecked:
            self.checked_num -= 1
        else:
            self.checked_num += 1
        if self.checked_num == 0:
            self.export_btn.setDisabled(True)
        else:
            self.export_btn.setEnabled(True)

    def preview(self):
        if self.list_view.currentItem() is None:
            return  # TODO check this
        name = str(self.list_view.currentItem().text())
        self.viewer.preview_object(self.export_dict[name])

    def check_change(self):
        item = self.list_view.currentItem()  # type: QListWidgetItem
        index = self.list_view.currentRow()
        checked = item.checkState() == Qt.CheckState.Checked
        self.check_state[index] = checked
        self.export_btn.setEnabled(np.any(self.check_state))

    def uncheck_all(self):
        for index in range(self.list_view.count()):
            item = self.list_view.item(index)
            item.setCheckState(Qt.CheckState.Unchecked)
        self.check_state[...] = False
        self.export_btn.setDisabled(True)
        self.checked_num = 0

    def check_all(self):
        for index in range(self.list_view.count()):
            item = self.list_view.item(index)
            item.setCheckState(Qt.CheckState.Checked)
        self.checked_num = len(self.export_dict)
        self.check_state[...] = True
        self.export_btn.setDisabled(False)

    def get_export_list(self):
        res = []
        for num in range(self.list_view.count()):
            item = self.list_view.item(num)
            if item.checkState() == Qt.CheckState.Checked:
                res.append(str(item.text()))
        return res

    def get_checked(self):
        return self.get_export_list()


class ImportDialog(QDialog):
    def __init__(
        self,
        import_dict: typing.Dict[str, typing.Any],
        local_dict: typing.Dict[str, typing.Any],
        viewer: typing.Type[ObjectPreviewProtocol],
        expected_type: typing.Optional[typing.Type] = None,
        parent: typing.Optional[QWidget] = None,
    ):
        """
        :param import_dict: dict with data to import
        :param local_dict: dict with data already in project
        :param viewer: class used to preview data
        :param expected_type: type of data to import
        :param parent: parent qt widget
        """
        super().__init__(parent=parent)
        self.setWindowTitle("Import")
        self.viewer = viewer()
        self.local_viewer = viewer()
        self.import_dict = import_dict
        self.local_dict = local_dict
        self.expected_type = expected_type
        conflicts = set(local_dict.keys()) & set(import_dict.keys())

        self.list_view = QTreeWidget()
        self.list_view.setColumnCount(4)
        self.radio_group_list = []
        self.checked_num = len(import_dict)

        ommit_count = 0

        for name in sorted(import_dict):
            if self.expected_type is not None and not isinstance(import_dict[name], self.expected_type):
                ommit_count += 1
                continue

            item = QTreeWidgetItem()
            item.setText(0, name)
            # noinspection PyTypeChecker
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(0, Qt.CheckState.Checked)
            self.list_view.addTopLevelItem(item)
            if name in conflicts:
                group = QButtonGroup()
                overwrite = QRadioButton("Overwrite")
                overwrite.setChecked(True)
                rename = QRadioButton("Rename")
                new_name = QLineEdit()
                new_name.textChanged.connect(self.block_import(rename, new_name))
                rename.toggled.connect(self.block_import(rename, new_name))
                overwrite.toggled.connect(self.block_import(rename, new_name))

                rename.toggled.connect(self.rename_func(name, new_name, rename))
                group.addButton(overwrite)
                group.addButton(rename)
                self.radio_group_list.append(group)
                self.list_view.setItemWidget(item, 1, overwrite)
                self.list_view.setItemWidget(item, 2, rename)
                self.list_view.setItemWidget(item, 3, new_name)

        self.import_btn = QPushButton("Import")
        self.cancel_btn = QPushButton("Cancel")
        self.check_btn = QPushButton("Check all")
        self.uncheck_btn = QPushButton("Uncheck all")

        self.cancel_btn.clicked.connect(self.close)
        self.import_btn.clicked.connect(self.accept)
        self.check_btn.clicked.connect(self.check_all)
        self.uncheck_btn.clicked.connect(self.uncheck_all)

        self.list_view.itemSelectionChanged.connect(self.preview)
        self.list_view.itemChanged.connect(self.checked_change)
        self.setup_ui(ommit_count)

    def setup_ui(self, ommit_count):
        layout = QVBoxLayout()
        info_layout = QHBoxLayout()
        info_layout.addWidget(self.list_view, 2)
        v1_lay = QVBoxLayout()
        v1_lay.addWidget(QLabel("Import:"))
        v1_lay.addWidget(self.viewer)
        info_layout.addLayout(v1_lay, 1)
        v2_lay = QVBoxLayout()
        v2_lay.addWidget(QLabel("Local:"))
        v2_lay.addWidget(self.local_viewer)
        info_layout.addLayout(v2_lay, 1)
        layout.addLayout(info_layout)
        if ommit_count:
            layout.addWidget(
                QLabel(f"Omitted {ommit_count} objects as they are not of type {self.expected_type.__name__}")
            )
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.check_btn)
        btn_layout.addWidget(self.uncheck_btn)
        btn_layout.addStretch()
        btn_layout.addWidget(self.import_btn)
        btn_layout.addWidget(self.cancel_btn)
        layout.addLayout(btn_layout)
        self.setLayout(layout)

    def rename_func(self, ob_name, new_name_field, rename_radio):
        end_reg = re.compile(r"(.*) \((\d+)\)$")

        def in_func():
            if not rename_radio.isChecked() or str(new_name_field.text()).strip():
                return

            match = end_reg.match(ob_name)
            if match:
                new_name_format = match[1] + " ({})"
                i = int(match[2]) + 1
            else:
                new_name_format = ob_name + " ({})"
                i = 1
            while new_name_format.format(i) in self.local_dict:
                i += 1
            new_name_field.setText(new_name_format.format(i))

        return in_func

    def block_import(self, radio_btn, name_field):
        def inner_func():
            text = str(name_field.text()).strip()
            self.import_btn.setDisabled(not text and radio_btn.isChecked())

        return inner_func

    def preview(self):
        item = self.list_view.currentItem()
        name = str(item.text(0))
        self.viewer.preview_object(self.import_dict[name])
        if self.list_view.itemWidget(item, 1) is not None:
            self.local_viewer.preview_object(self.local_dict[name])
        else:
            self.local_viewer.clear()

    def checked_change(self, item, _):
        if item.checkState(0) == Qt.CheckState.Unchecked:
            self.checked_num -= 1
            if self.list_view.itemWidget(item, 1) is not None:
                self.list_view.itemWidget(item, 1).setDisabled(True)
                self.list_view.itemWidget(item, 2).setDisabled(True)
                self.list_view.itemWidget(item, 3).setDisabled(True)
        else:
            self.checked_num += 1
            if self.list_view.itemWidget(item, 1) is not None:
                self.list_view.itemWidget(item, 1).setEnabled(True)
                self.list_view.itemWidget(item, 2).setEnabled(True)
                self.list_view.itemWidget(item, 3).setEnabled(True)
        if self.checked_num == 0:
            self.import_btn.setDisabled(True)
        else:
            self.import_btn.setEnabled(True)

    def get_import_list(self):
        res = []
        for index in range(self.list_view.topLevelItemCount()):
            item = self.list_view.topLevelItem(index)
            if item.checkState(0) == Qt.CheckState.Checked:
                chk = self.list_view.itemWidget(item, 2)
                if chk is not None and typing.cast(QRadioButton, chk).isChecked():
                    res.append((item.text(0), typing.cast(QLineEdit, self.list_view.itemWidget(item, 3)).text()))
                else:
                    name = str(item.text(0))
                    res.append((name, name))
        return res

    def uncheck_all(self):
        for index in range(self.list_view.topLevelItemCount()):
            item = self.list_view.topLevelItem(index)
            item.setCheckState(0, Qt.CheckState.Unchecked)
        self.import_btn.setDisabled(True)
        self.checked_num = 0

    def check_all(self):
        for index in range(self.list_view.topLevelItemCount()):
            item = self.list_view.topLevelItem(index)
            item.setCheckState(0, Qt.CheckState.Checked)
        self.checked_num = len(self.import_dict)
        self.import_btn.setDisabled(False)



import json

from PartSegCore import Units
from PartSegCore.image_operations import RadiusType
from PartSegCore.io_utils import find_problematic_leafs
from PartSegCore.json_hooks import PartSegEncoder, partseg_object_hook
from PartSegCore.utils import ProfileDict


def test_profile_hook_colormap_load(bundle_test_dir):
    with open(bundle_test_dir / "view_settings_v0.12.6.json") as f_p:
        res = json.load(f_p, object_hook=partseg_object_hook)

    assert isinstance(res, ProfileDict)
    assert res.verify_data()


def test_dump_custom_types():
    prof_dict = ProfileDict()
    prof_dict.set("a.b.c", 1)
    data = {"a": RadiusType.R2D, "b": prof_dict}
    text = json.dumps(data, cls=PartSegEncoder)
    loaded = json.loads(text, object_hook=partseg_object_hook)
    assert loaded["a"] == RadiusType.R2D
    assert loaded["b"].get("a.b.c") == 1


def test_error_reported(bundle_test_dir):
    with open(bundle_test_dir / "problematic_profile_dict.json") as f_p:
        res = json.load(f_p, object_hook=partseg_object_hook)

    assert isinstance(res, ProfileDict)
    assert not res.verify_data()
    error_data = res.pop_errors()
    assert len(error_data) == 2
    assert len(find_problematic_leafs(error_data[0][1])) == 1
    assert len(find_problematic_leafs(error_data[1][1])) == 2


def test_plugin_bugfix():
    data = '{"__class__": "plugins.PartSegCore.mask_partition_utils.BorderRimParameters", "__class_version_dkt__": {"plugins.PartSegCore.mask_partition_utils.BorderRimParameters": "0.0.0", "PartSegCore.utils.BaseModel": "0.0.0"}, "__values__": {"distance": 500, "units": {"__class__": "PartSegCore.universal_const.Units", "__class_version_dkt__": {"plugins.PartSegCore.universal_const.Units": "0.0.0"}, "__values__": {"value": 2}}}}'  # noqa: E501
    res = json.loads(data, object_hook=partseg_object_hook)
    assert res.distance == 500
    assert res.units == Units.nm


def _test_prepare(bundle_test_dir):  # pragma: no cover
    """Prepare test data for test_error_reported. After call this the names of class should be manually broken"""
    from PartSegCore.segmentation.noise_filtering import BilateralNoiseFilteringParams, GaussNoiseFilteringParams

    dkt = ProfileDict()
    dkt.set("a.b", GaussNoiseFilteringParams())
    dkt.set("a.c", BilateralNoiseFilteringParams())
    dkt.set("b.a.d", GaussNoiseFilteringParams())
    dkt.set("b.a.e", BilateralNoiseFilteringParams())
    dkt.set("b.a.f", GaussNoiseFilteringParams())

    with open(bundle_test_dir / "problematic_profile_dict.json", "w") as f_p:
        json.dump(dkt, f_p, cls=PartSegEncoder)



# pylint: disable=no-self-use
import os
import platform
from functools import partial

import pytest
from qtpy.QtCore import QCoreApplication

from PartSeg._launcher.main_window import MainWindow as LauncherMainWindow
from PartSeg._launcher.main_window import PartSegGUILauncher
from PartSeg._launcher.main_window import Prepare as LauncherPrepare
from PartSeg._roi_analysis import main_window as analysis_main_window
from PartSeg._roi_mask import main_window as mask_main_window

GITHUB_ACTIONS = "GITHUB_ACTIONS" in os.environ


def empty(*_):
    """To silent some functions"""


class TestLauncherMainWindow:
    def test_opening(self, qtbot):
        main_window = LauncherMainWindow("Launcher")
        qtbot.addWidget(main_window)

    # @pytest.mark.skipif((platform.system() == "Linux") and CI_BUILD, reason="vispy problem")
    @pytest.mark.enablethread()
    @pytest.mark.pyside_skip()
    @pytest.mark.windows_ci_skip()
    def test_open_mask(self, qtbot, monkeypatch, tmp_path):
        monkeypatch.setattr(mask_main_window, "CONFIG_FOLDER", str(tmp_path))
        if platform.system() == "Linux" and GITHUB_ACTIONS:
            monkeypatch.setattr(mask_main_window.MainWindow, "show", empty)
        main_window = PartSegGUILauncher()
        qtbot.addWidget(main_window)
        with qtbot.waitSignal(main_window.prepare.finished, timeout=10**4):
            main_window.launch_mask()
        QCoreApplication.instance().processEvents()
        qtbot.add_widget(main_window.wind[0])
        main_window.wind[0].hide()
        qtbot.wait(50)

    # @pytest.mark.skipif((platform.system() == "Linux") and CI_BUILD, reason="vispy problem")
    @pytest.mark.enablethread()
    @pytest.mark.windows_ci_skip()
    @pytest.mark.pyside_skip()
    def test_open_analysis(self, qtbot, monkeypatch, tmp_path):
        monkeypatch.setattr(analysis_main_window, "CONFIG_FOLDER", str(tmp_path))
        if platform.system() in {"Darwin", "Linux"} and GITHUB_ACTIONS:
            monkeypatch.setattr(analysis_main_window.MainWindow, "show", empty)
        main_window = PartSegGUILauncher()
        qtbot.addWidget(main_window)
        with qtbot.waitSignal(main_window.prepare.finished):
            main_window.launch_analysis()
        QCoreApplication.processEvents()
        qtbot.wait(50)
        qtbot.add_widget(main_window.wind[0])
        main_window.wind[0].hide()
        qtbot.wait(50)

    def test_prepare(self):
        prepare = LauncherPrepare("PartSeg._roi_analysis.main_window")
        prepare.run()
        assert isinstance(prepare.result, partial)
        assert prepare.result.func is analysis_main_window.MainWindow



from enum import Enum
from typing import Iterable, List, Union

import numpy as np
import SimpleITK as sitk
from local_migrator import register_class


@register_class
class RadiusType(Enum):
    """
    If operation should be performed and if on each layer separately on globally
    """

    NO = 0  #: No operation
    R2D = 1  #: operation in each layer separately
    R3D = 2  #: operation on whole stack

    def __str__(self):
        return ["No", "2d", "3d"][self.value]


@register_class
class NoiseFilterType(Enum):
    No = 0
    Gauss = 1
    Median = 2
    Bilateral = 3

    def __str__(self):
        return self.name


def _generic_image_operation(image, radius, fun, layer):
    if image.ndim == 3 and image.shape[0] == 1:
        layer = True
    if image.ndim == 2:
        layer = False
    if image.dtype == bool:
        image = image.astype(np.uint8)
    if isinstance(radius, (list, tuple)):
        radius = list(reversed(radius))
    if not layer and image.ndim <= 3:
        return sitk.GetArrayFromImage(fun(sitk.GetImageFromArray(image), radius))
    return _generic_image_operations_recurse(np.copy(image), radius, fun, layer)


def _generic_image_operations_recurse(image, radius, fun, layer):
    if (not layer and image.ndim == 3) or image.ndim == 2:
        return sitk.GetArrayFromImage(fun(sitk.GetImageFromArray(image), radius))

    for layer_data in image:
        layer_data[...] = _generic_image_operations_recurse(layer_data, radius, fun, layer)
    return image


def gaussian(image: np.ndarray, radius: float, layer=True):
    """
    Gaussian blur of image.

    :param np.ndarray image: image to apply gaussian filter
    :param float radius: radius for gaussian kernel
    :param bool layer: if operation should be run on each layer separately
    :return:
    """
    return _generic_image_operation(image, radius, sitk.DiscreteGaussian, layer)


def bilateral(image: np.ndarray, radius: float, layer=True):
    """
    Gaussian blur of image.

    :param np.ndarray image: image to apply gaussian filter
    :param float radius: radius for gaussian kernel
    :param bool layer: if operation should be run on each layer separately
    :return:
    """
    return _generic_image_operation(image, radius, sitk.Bilateral, layer)


def median(image: np.ndarray, radius: Union[int, List[int]], layer=True):
    """
    Median blur of image.

    :param np.ndarray image: image to apply median filter
    :param float radius: radius for median kernel
    :param bool layer: if operation should be run on each layer separately
    :return:
    """
    if not isinstance(radius, Iterable):
        radius = [radius] * min(image.ndim, 2 if layer else 3)
    return _generic_image_operation(image, radius, sitk.Median, layer)


def dilate(image, radius, layer=True):
    """
    Dilate of image.

    :param image: image to apply dilation
    :param radius: dilation radius
    :param layer: if operation should be run on each layer separately
    :return:
    """
    return _generic_image_operation(image, radius, sitk.GrayscaleDilate, layer)


def apply_filter(filter_type, image, radius, layer=True) -> np.ndarray:
    """
    Apply operation selected by filter type to image.

    :param NoiseFilterType filter_type:
    :param np.ndarray image:
    :param float radius:
    :param bool layer:
    :return: image after operation
    :rtype: np.ndarray
    """
    if filter_type == NoiseFilterType.Gauss:
        return gaussian(image, radius, layer)
    if filter_type == NoiseFilterType.Median:
        return median(image, int(radius), layer)
    return image


def erode(image, radius, layer=True):
    """
    Erosion of image

    :param image: image to apply erosion
    :param radius: erosion radius
    :param layer: if operation should be run on each layer separately
    :return:
    """
    return _generic_image_operation(image, radius, sitk.GrayscaleErode, layer)


def to_binary_image(image):
    """Convert image to binary. All positive values are set to 1."""
    return np.array(image > 0).astype(np.uint8)



from typing import Dict, NamedTuple, Optional

from qtpy.QtWidgets import QComboBox, QDialog, QGridLayout, QPushButton, QStackedWidget

from PartSeg.common_gui.algorithms_description import FormWidget
from PartSegCore.image_transforming import TransformBase, image_transform_dict
from PartSegImage import Image


class ImageAdjustTuple(NamedTuple):
    values: dict
    algorithm: TransformBase


class ImageAdjustmentDialog(QDialog):
    def __init__(self, image: Image, transform_dict: Optional[Dict[str, TransformBase]] = None):
        super().__init__()
        if transform_dict is None:
            transform_dict = image_transform_dict
        self.choose = QComboBox()
        self.stacked = QStackedWidget()
        for key, val in transform_dict.items():
            self.choose.addItem(key)
            initial_values = val.calculate_initial(image)
            form_widget = FormWidget(val.get_fields_per_dimension(list(image.get_dimension_letters())), initial_values)
            self.stacked.addWidget(form_widget)

        self.choose.currentIndexChanged.connect(self.stacked.setCurrentIndex)

        self.cancel_btn = QPushButton("Cancel")
        self.cancel_btn.clicked.connect(self.reject)

        self.process_btn = QPushButton("Process")
        self.process_btn.clicked.connect(self.process)
        self.transform_dict = transform_dict
        self.result_val: ImageAdjustTuple = None

        layout = QGridLayout()
        layout.addWidget(self.choose, 0, 0, 1, 3)
        layout.addWidget(self.stacked, 1, 0, 1, 3)
        layout.addWidget(self.cancel_btn, 2, 0)
        layout.addWidget(self.process_btn, 2, 2)
        self.setLayout(layout)
        self.setWindowTitle("Image adjustment")

    def process(self):
        values = self.stacked.currentWidget().get_values()
        algorithm = self.transform_dict[self.choose.currentText()]
        self.result_val = ImageAdjustTuple(values, algorithm)

        self.accept()



import typing

import numpy as np

from PartSegCore.analysis import ProjectTuple
from PartSegCore.io_utils import LoadBase, WrongFileTypeException
from PartSegCore.mask.io_functions import MaskProjectTuple


def _image_to_layers(project_info, scale, translate):
    res_layers = []
    if project_info.image.name == "ROI" and project_info.image.channels == 1:
        res_layers.append(
            (
                project_info.image.get_channel(0),
                {"scale": scale, "name": project_info.image.channel_names[0], "translate": translate},
                "labels",
            )
        )
    else:
        res_layers.extend(
            (
                project_info.image.get_channel(i),
                {
                    "scale": scale,
                    "name": project_info.image.channel_names[i],
                    "blending": "additive",
                    "translate": translate,
                },
                "image",
            )
            for i in range(project_info.image.channels)
        )

    return res_layers


def project_to_layers(project_info: typing.Union[ProjectTuple, MaskProjectTuple]):
    res_layers = []
    if project_info.image is not None and not isinstance(project_info.image, str):
        scale = project_info.image.normalized_scaling()
        translate = project_info.image.shift
        translate = (0,) * (len(project_info.image.axis_order.replace("C", "")) - len(translate)) + translate
        res_layers.extend(_image_to_layers(project_info, scale, translate))
        if project_info.roi_info.roi is not None:
            res_layers.append(
                (
                    project_info.image.fit_array_to_image(project_info.roi_info.roi),
                    {"scale": scale, "name": "ROI", "translate": translate},
                    "labels",
                )
            )
        if project_info.roi_info.alternative:
            res_layers.extend(
                (
                    project_info.image.fit_array_to_image(roi),
                    {
                        "scale": scale,
                        "name": name,
                        "translate": translate,
                        "visible": False,
                    },
                    "labels",
                )
                for name, roi in project_info.roi_info.alternative.items()
            )

        if project_info.mask is not None:
            res_layers.append(
                (
                    project_info.image.fit_array_to_image(project_info.mask),
                    {"scale": scale, "name": "Mask", "translate": translate},
                    "labels",
                )
            )
    else:
        if isinstance(project_info, MaskProjectTuple) and project_info.spacing is not None:
            scale = np.multiply(project_info.spacing, 10**9)
        else:
            scale = None
        if project_info.roi_info.roi is not None:
            res_layers.append(
                (
                    project_info.roi_info.roi,
                    {"scale": scale, "name": "ROI"},
                    "labels",
                )
            )
    return res_layers


def partseg_loader(loader: typing.Type[LoadBase], path: str):
    load_locations = [path]
    load_locations.extend(loader.get_next_file(load_locations) for _ in range(1, loader.number_of_files()))

    try:
        project_info = loader.load(load_locations)
    except WrongFileTypeException:  # pragma: no cover
        return None

    if isinstance(project_info, (ProjectTuple, MaskProjectTuple)):
        return project_to_layers(project_info)
    return None  # pragma: no cover
